(******************************************************************************

  Modeling and Verifying Properties of Security Protocols in Wireless Networks

  Module: distance/System.thy (Isabelle/HOL 2009)

  Copyright (C) 2007-2009 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
  License: BSD (see COPYRIGHT)

******************************************************************************)

header {* Inductively defined Systems parameterized by Protocols *}

theory System imports Distance MessageDerivation begin

subsection {* Protocol independent Facts *}

fun
  maxtime :: "'msg trace \<Rightarrow> time"
 where
   "maxtime [] = (0::real)"
 | "maxtime (x#xs) = max (fst x) (maxtime xs)"

text {* case distinction needed for some proofs *}

lemma set_two_elem_cases:
  assumes trxa: "eva \<in> set (x#tr)" and trxb:  "evb \<in> set (x#tr)"
  assumes ina_inb: "\<lbrakk> eva \<in> set tr; evb \<in> set tr \<rbrakk> \<Longrightarrow> P tr eva evb x"
  assumes ina_eqb: "\<lbrakk> eva \<in> set tr; evb = x     ; eva \<noteq> x \<rbrakk> \<Longrightarrow> P tr eva evb x"
  assumes eqa_inb: "\<lbrakk> eva = x     ; evb \<in> set tr; evb \<noteq> x \<rbrakk> \<Longrightarrow> P tr eva evb x"
  assumes eqa_eqb: "\<lbrakk> eva = x     ; evb = x  \<rbrakk> \<Longrightarrow> P tr eva evb x"
  shows "P tr eva evb x"
proof cases
  assume ina: "eva \<in> set tr"
  show ?thesis
  proof cases
    assume "evb \<in> set tr"
    from this ina ina_inb show ?thesis by auto
  next
    assume "evb \<notin> set tr"
    from this trxb have eqb: "evb = x" by auto
    show ?thesis
    proof cases
      assume "eva = x"
      from this eqb eqa_eqb show ?thesis by auto
    next
      assume "eva \<noteq> x"
      from this ina_eqb eqb ina show ?thesis by auto
    qed
  qed
next
  assume "eva \<notin> set tr"
  from this trxa have eqa: "eva = x" by auto
  show ?thesis
  proof cases
    assume "evb \<notin> set tr"
    from this trxb have "evb = x" by auto
    from this eqa eqa_eqb show ?thesis by auto
  next
    assume "\<not> evb \<notin> set tr" -- "ugly"
    then have inb: "evb \<in> set tr" by simp
    show ?thesis
    proof cases
      assume "evb = x"
      from this eqa eqa_eqb show ?thesis by auto
    next
      assume "evb \<noteq> x"
      from this eqa_inb eqa inb show ?thesis by auto
    qed
  qed
qed

fun
  beforeEvent :: "[(time * 'msg event), 'msg trace] \<Rightarrow> 'msg trace"
 where
  "beforeEvent e (x#xs) = (if x = e \<and> (e \<notin> set xs) then xs else beforeEvent e xs)" |
  "beforeEvent e [] = []"

lemma beforeEvent_Send_Recv [simp]:
  "beforeEvent (ta, Send A ma L) ((tb, Recv B mb) # tra)
  = beforeEvent (ta, Send A ma L) (tra)"
by (auto simp add: beforeEvent.simps)

lemma beforeEvent_Send_Claim [simp]:
  "beforeEvent (ta, Send A ma L) ((tb, Claim B mb) # tra)
  = beforeEvent (ta, Send A ma L) (tra)"
by (auto simp add: beforeEvent.simps)

lemma beforeEvent_Send_other [simp]:
 "\<lbrakk> ma \<noteq> mb \<rbrakk>
 \<Longrightarrow> beforeEvent (ta, Send A ma La) ((tb, Send B mb Lb) # tra) = beforeEvent (ta, Send A ma La) tra"
  apply (auto simp add: beforeEvent.simps)
done

lemma beforeEvent_send_other2 [simp]:
 "\<lbrakk>  ta = tb \<longrightarrow> A = B \<longrightarrow> La = Lb \<longrightarrow> ma \<noteq> mb \<rbrakk>
 \<Longrightarrow> beforeEvent (ta, Send A ma La) ((tb, Send B mb Lb) # tra) = beforeEvent (ta, Send A ma La) tra"
  apply (auto simp add: beforeEvent.simps)
done


lemma beforeEvent_same [simp]:
  "e \<notin> set tr \<Longrightarrow> beforeEvent e (e # tr) = tr"
  apply (auto simp add: beforeEvent.simps)
done

subsubsection {* Simplification rules for the used Set and beforeEvent *}

lemma (in MESSAGE_DERIVATION) used_beforeEvent:
  "X \<notin> used evs \<Longrightarrow>  X \<notin> used (beforeEvent ev evs)"
proof (induct evs rule: trace_induct)
  case 1 thus ?case by auto
next
  case (2 t ev evs) thus ?case by (auto split: event.split_asm)
qed

lemma beforeEvent_subset:
  "x \<in> set (beforeEvent y xs) \<Longrightarrow> x \<in> set xs"
  apply (induct xs, auto split: split_if_asm)
done

lemma (in INITSTATE) fresh_mono[intro]:
  "m \<notin> usedI (beforeEvent e (x#tr)) \<Longrightarrow> m \<notin> usedI (beforeEvent e tr)"
  apply (auto simp add: usedI_def split: split_if_asm)
  apply (drule Used_imp_send_parts)
  apply (elim exE conjE)
  apply (drule beforeEvent_subset)
  apply (drule Send_imp_parts_used)
by auto

text {* time increases monotonically in traces *}

lemma maxtime_non_negative [intro, simp]:
  "maxtime l >= 0"
proof (induct l rule: trace_induct)
  case 1 show ?case by auto
next
  case 2 thus ?case by auto
qed

lemma maxtime_geq_elem:
  assumes "maxtime tr \<le> t" and "(t', ev) \<in> set tr"
  shows   "t' \<le> t" using prems
proof (induct tr rule: trace_induct)
  case 1 thus ?case by auto
next
  case (2 t ev tr)
  thus ?case by (auto)
qed

subsection {* Protocols and the parameterized System Definition *}

types
  friendid = nat (* move *)
  transmitterid = nat
  receveiverid = nat

text {* "clocktime A t" returns the time of agent A's clock at time t *}

consts
  clocktime :: "friendid \<Rightarrow> time \<Rightarrow> time"

fun
  occursAt :: "'msg event \<Rightarrow> agent"
 where
    "occursAt (Send (Tx A i) m L) = A"
  | "occursAt (Recv (Rx A i) m) = A"
  | "occursAt (Claim A m)        = A"

definition
  view :: "[friendid, 'msg trace] \<Rightarrow> 'msg trace"
 where
  "view A tr = [(clocktime A t,ev) . (t, ev) \<leftarrow> tr, occursAt ev = (Honest A)]"

lemma view_occurs_at:
  "(t,ev) \<in> set (view A tr) \<Longrightarrow> occursAt ev = (Honest A)"
by (auto split: event.split split_if_asm simp add: occursAt.simps view_def)

lemma view_subset:
  "snd`(set (view A tr)) \<subseteq> snd`(set tr)"
apply (auto simp add: view_def split: split_if_asm)
by force

lemma (in INITSTATE) used_view_subset:
  "used (view A tr) \<subseteq> used tr"
apply (induct tr)
apply (force simp add: view_def used.simps split: split_if_asm)
apply (auto simp add: used.simps split: event.split)
apply (auto simp add: view_def split: split_if_asm)
done

lemma (in INITSTATE_NONONCE) Used_imp_subterm_Send:
  assumes u: "Nonce A NA \<in> used tr"
  shows a: "\<exists>t B i X L. (t, Send (Tx B i) X L) \<in> set tr \<and> Nonce A NA \<in> subterms {X}" using u
proof (induct tr rule: trace_induct)
  case 1 thus ?case by (auto simp add: used.simps)
next
  case (2 ts x xs)
  with prems show ?case
    apply (auto simp add: used.simps split: split_if_asm event.split_asm)
    apply (rule_tac x=ts in exI,case_tac transmitter)
    apply (rule_tac x="agent" in exI, rule_tac x=nat in exI)
    apply (rule_tac x=msg in exI)
    apply auto
    apply (rule_tac x=t in exI)
    apply (rule_tac x=B in exI)
    apply (rule_tac x=i in exI)
    apply (rule_tac x=X in exI)
    by auto
qed

text {* protocols return protoEvents to ensure that protocols only create events for the
        agent running the protocol *}

datatype 'msg protoEv = SendEv transmitterid "'msg list" | ClaimEv

text {* a protocol step returns the set of events that can be executed by the agent
        executing the step *}

types
  'msg step = "['msg trace, friendid,time] \<Rightarrow> ('msg * 'msg protoEv) set"
  'msg proto = "('msg step) set"

fun
   createEv :: "[friendid,'msg protoEv,'msg] \<Rightarrow> 'msg event"
 where
  "createEv fid (SendEv txid L) m = Send (Tx (Honest fid) txid) m L"
| "createEv fid ClaimEv m = Claim (Honest fid) m"

text {*Construct the set of possible events (following the rules of the protocol) as a
       set of events, for a given trace tr*}

locale INITSTATE_DM = MESSAGE_THEORY_DM  + INITSTATE

locale PROTOCOL = INITSTATE_DM _ _ _ _ _ _ _ Key for Key :: "nat \<Rightarrow> 'msg" + 
  fixes proto :: "'msg proto"

inductive_set (in PROTOCOL)
  sys :: "'msg trace set"
 where
   Nil [intro] : "[] \<in> sys"
 | Fake: 
   "\<lbrakk> tr \<in> sys;  t >=  maxtime tr;
      X \<in> DM (Intruder I) (knowsI (Intruder I) tr) \<rbrakk>
   \<Longrightarrow> (t, Send (Tx (Intruder I) j) X [])  # tr \<in> sys"

 | Con : 
   "\<lbrakk> tr \<in> sys; trecv >= maxtime tr; 
      (tsend, Send (Tx A i) m L) \<in> set tr;
      cdistM (Tx A i) (Rx B j) = Some tab;
      trecv \<ge> tsend + tab  \<rbrakk>
    \<Longrightarrow> (trecv, Recv (Rx B j) m) # tr \<in> sys"

 | Proto :
   "\<lbrakk> tr \<in> sys; t >=  maxtime tr;
      step \<in> proto; (m,pEv) \<in> step (view A tr) A (clocktime A t);
      m \<in> DM (Honest A) (knowsI (Honest A) tr) \<rbrakk>
    \<Longrightarrow> ((t,createEv A pEv m)#tr) \<in> sys"


text {* default transmitter/receiver *}

syntax
  Tr :: "agent \<Rightarrow> transmitter"
  Rec :: "agent \<Rightarrow> receiver"

translations
  "Tr A" == "Tx A 0"
  "Rec A" == "Rx A 0"

syntax
  Tu :: "agent \<Rightarrow> transmitter"
  Ru :: "agent \<Rightarrow> receiver"

text {* each agent has only one ultrasound receiver and transmitter *}

translations
  "Tu A" == "Tx A 1"
  "Ru A" == "Rx A 1"

end