(******************************************************************************

  Modeling and Verifying Properties of Security Protocols in Wireless Networks

  Module: distance/SystemOrigination.thy (Isabelle/HOL 2009)

  Copyright (C) 2007-2009 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
  License: BSD (see COPYRIGHT)

******************************************************************************)

theory SystemOrigination imports SystemSimps begin

definition
  messagesProtoTrHonest :: "['msg proto,'msg trace,friendid,time] \<Rightarrow> 'msg set" where
  "messagesProtoTrHonest proto tr fid t ==
     fst`(Union ((\<lambda>step. step (view fid tr) fid t)`proto))"

definition
  messagesProto :: "['msg proto] \<Rightarrow> 'msg set" where
  "messagesProto proto == (UN tr fid t. messagesProtoTrHonest proto tr fid t)"

definition
  messagesProtoTr :: "['msg proto,'msg trace] \<Rightarrow> 'msg set" where
  "messagesProtoTr proto tr == (UN fid t. messagesProtoTrHonest proto tr fid t)"

lemmas messagesProtoDefs = messagesProto_def messagesProtoTrHonest_def
                           messagesProtoTr_def

subsection {* Signature Creation and Key Knowledge by Dishonest Users *}

locale PROTOCOL_SYMKEYS_NOKEYS = PROTOCOL + INITSTATE_SYMKEYS +
  assumes protoSendNoKeys:
    "!!A B tr. Key (symKey A B) \<in> parts (messagesProtoTr proto tr) \<Longrightarrow>
        \<exists> C t i M. (t, Recv (Rx C i) M) \<in> set tr \<and> Key (symKey A B) \<in> parts {M}"

locale PROTOCOL_PKSIG_NOKEYS = PROTOCOL + INITSTATE_PKSIG +
  assumes protoSendNoKeys:
    "!!B. Key (priSK (Honest B)) \<notin> parts (messagesProto proto)"

lemma (in PROTOCOL_PKSIG_NOKEYS) no_keys_send_honest:
  assumes sys_proto: "tr \<in> sys" and
          msg: "(t, Send (Tx (Honest A) i) X L) \<in> set tr"
  shows "Key (priSK (Honest B)) \<notin> parts {X}"
  using sys_proto msg
  apply (induct tr)
  apply (insert protoSendNoKeys)
  apply auto
  apply (subgoal_tac "X=m") defer apply (case_tac pEv, force, force)
  apply (subgoal_tac "X \<in> parts (messagesProto proto)")
  apply (subgoal_tac "Key (priSK (Honest B)) \<in> parts (messagesProto proto)")
  apply (force) apply (rule parts.elem_trans) apply (auto)
  apply (case_tac pEv)
  apply auto
  apply (auto simp add: messagesProtoDefs)
  apply (rule_tac x=tr in exI, rule_tac x=A in exI, rule_tac x="clocktime A t" in exI,
         rule_tac x=step in bexI)
  apply (auto)
  apply (subgoal_tac "{X} \<subseteq> (fst ` step (view A tr) A (clocktime A t))")
by force+

(* here, we need a separate lemmas that states that
   B /= A cannot derive a key of A if its not already
   in parts. *)
lemma (in PROTOCOL_PKSIG_NOKEYS) keys_not_send_received:
  assumes rang: "tr \<in> sys" and
          sr: "(tsend, Send (Tx A i) X L) \<in> set tr \<or> (trecv, Recv (Rx A i) X) \<in> set tr"
  shows "Key (priSK (Honest B)) \<notin> parts {X}"
  using rang sr
proof (induct tr arbitrary: A B X i L tsend trecv rule: sys.induct)
  case Nil
  thus ?case by auto
next
  case (Fake tr tintr mintr I j)
  let ?x   = "(tintr, Send (Tx (Intruder I) j) mintr [])" and
      ?eva = "(tsend, Send (Tx A i) X L)" and
      ?evb = "(trecv, Recv (Rx A i) X)"
  show ?case
  proof cases
    assume "?evb \<in> set (?x # tr)"
    hence "?evb \<in> set tr" by auto
    with prems show ?case apply - apply (rule Fake.hyps(2)) by (auto)
  next
    assume "\<not> (?evb \<in> set (?x # tr))"
    with `?eva \<in> set (?x # tr) \<or> ?evb \<in> set (?x # tr)`
      have "?eva \<in> set (?x # tr)" by auto
    show ?case
    proof cases
      assume "?x=?eva"
      hence "X=mintr" and "Intruder I=A" and "i=j" by auto
      hence xdy: "X \<in> DM A (knowsI A tr)" using prems by auto
      show ?case
      proof cases
        assume "Key (priSK (Honest B)) \<in> parts {X}"
        hence ex: "\<exists> Z \<in> knowsI A tr. Key (priSK (Honest B)) \<in> parts {Z}"
	  using xdy `Intruder I=A` apply -
	  apply (subgoal_tac "Key (priSK (Honest B)) \<in> parts (DM A (knowsI A tr))")
	  apply (drule key_parts_DM_key)
	  apply (drule parts.singleton) back
	  apply auto
	  apply (subgoal_tac "{X} \<subseteq> (DM (Intruder I) (knowsI (Intruder I) tr))")
	  apply (drule parts.mono)
	  apply (erule subsetD)
	  by auto
	then obtain Z where knowsIz: "Z \<in> knowsI A tr"
	                and partsz: " Key (priSK (Honest B)) \<in> parts {Z}" by auto
        have ex: "(\<exists>t i. (t, Recv (Rx A i) Z) \<in> set tr) \<or> Z \<in> initState A"
          using knowsIz apply - apply (rule knowsI_A_imp_Recv_initState) by auto
        from partsz `Intruder I=A` have "Z \<notin> initState A" apply auto
          apply (drule_tac H="initState (Intruder I)" in parts.trans) apply force
	  apply (drule parts_subset_subterms[THEN subsetD])
          apply (auto simp add: priSK_notknown_other_subterms)
	  done
        with ex have "\<exists>t k. (t, Recv (Rx A k) Z) \<in> set tr" by auto
        then obtain t k where recvtr: "(t, Recv (Rx A k) Z) \<in> set tr" by auto
        with prems have  "Key (priSK (Honest B)) \<notin> parts {Z}"
          apply - apply (rule Fake.hyps(2)) by auto
        with partsz show ?thesis by contradiction
      qed
    next
      assume "?x\<noteq>?eva"
      with `?eva \<in> set (?x#tr)` have "?eva \<in> set tr" by auto
      with prems show ?case apply - apply (rule Fake.hyps(2)) by auto
    qed
  qed
next
  case  (Con tr tcrecv tsend C i Y LC D j dCD)
  thm prems
  let ?x = "(tcrecv, Recv (Rx D j) Y)" and
      ?eva="(tsend, Send (Tx A i) X L)" and ?evb="(trecv, Recv (Rx A i) X)"
  show ?case
  proof cases
    assume "?eva \<in> set (?x # tr)"
    hence "?eva \<in> set tr" by auto
    thus "Key (priSK (Honest B)) \<notin> parts {X}" apply -
      apply (rule_tac X=X in prems(5)) by auto
  next
    assume "\<not> (?eva \<in> set (?x # tr))"
    thm prems
    with `?eva \<in> set (?x # tr) \<or> ?evb \<in> set (?x # tr)` have "?evb \<in> set (?x # tr)"
      by auto
    show ?case
    proof cases
      assume "?x=?evb"
      hence "Y=X" by auto
      with prems have "?x#tr \<in> sys" apply - apply (rule sys.Con) by (auto)
      with `?x=?evb` have "?evb#tr \<in> sys" by auto
      with prems
        have "\<exists>C k tcsend Lc. (tcsend, Send (Tx C k) X Lc) \<in> set (?x#tr)
              \<and> trecv >= cdist (Tx C k) (Rx A i) \<and> cdistM (Tx C k) (Rx A i) \<noteq> None \<and>
                tcsend \<le> trecv - cdist (Tx C k) (Rx A i)"
      by (intro send_before_recv) (auto)
      then obtain C k tcsend Lc
        where   "(tcsend, Send (Tx C k) X Lc) \<in> set (?x#tr)" by auto
      hence p1: "(tcsend, Send (Tx C k) X Lc) \<in> set tr" by auto
      thus ?case apply - apply (rule prems(5)) by auto
    next
      assume "?x\<noteq>?evb"
      with `?evb \<in> set (?x#tr)` have "?evb \<in> set tr" by auto
      thus ?case apply - apply (rule prems(5)) by auto
    qed
  qed
next
  case (Proto tr t' step m pEv A')
  let ?x = "(t', createEv A' pEv m)" and
      ?eva="(tsend, Send (Tx A i) X L)" and ?evb="(trecv, Recv (Rx A i) X)"
  show ?case
  proof cases -- {* Recv event already in prefix of the trace, use IH *}
    assume "?evb \<in> set (?x # tr)"
    hence "?evb \<in> set tr" by (auto simp add: createEv.simps)
    thus ?case
      apply - apply (rule Proto.hyps(2) [where X=X and B=B]) by auto
  next -- {* Send event in trace *}
    assume "\<not> (?evb \<in> set (?x # tr))"
    with `?eva \<in> set (?x # tr) \<or> ?evb \<in> set (?x # tr)` have "?eva \<in> set (?x # tr)"
      by auto
    show ?case
    proof (cases pEv)
      fix tid list assume eveq: "pEv = SendEv tid list"
      show ?case
      proof cases
        assume eqeva: "?eva=?x"
	  -- {* send event added by Proto rule, use proto\_nokeys*}
        with prems eqeva have "?x#tr \<in> sys" apply - apply (rule sys.Proto)
          by (auto)
        moreover
        with eqeva have "?eva \<in> set (?x#tr)" by auto
        ultimately show ?case using eqeva
          apply - apply (rule_tac tr="?x # tr" in no_keys_send_honest)
	  apply assumption
	  apply (case_tac pEv)
	  apply auto
	  done
      next
        assume "?eva\<noteq>?x"
        with `?eva \<in> set (?x # tr)` have "?eva \<in> set tr" by auto
        with Proto.hyps(2) [where X=X and B=B and L=L] show ?case by (auto)
      qed
    next
      assume "pEv = ClaimEv"
      hence "?x\<noteq>?eva" by auto
      with `?eva \<in> set (?x#tr)` have "?eva \<in> set tr" by auto
      with Proto.hyps(2) [where X=X and B=B and L=L] show ?case by (auto)
    qed
  qed
qed

lemma tuple_fst_elem:
  "(a,b) \<in> H \<Longrightarrow> a \<in> fst`H"
  apply (auto simp add: image_def)
  apply (rule_tac x="(a,b)" in bexI)
  apply auto
done

lemma (in PROTOCOL_SYMKEYS_NOKEYS) keys_not_send_received:
  assumes rang: "tr \<in> sys" and
          sr: "(tsend, Send (Tx A i) X L) \<in> set tr \<or> (trecv, Recv (Rx A i) X) \<in> set tr"
  shows "Key (symKey (Honest B) (Honest C)) \<notin> parts {X}"
  using rang sr
proof (induct tr arbitrary: A B X i L tsend trecv rule: sys.induct)
  case Nil
  thus ?case by auto
next
  case (Fake tr tintr mintr I j)
  let ?x   = "(tintr, Send (Tx (Intruder I) j) mintr [])" and
      ?eva = "(tsend, Send (Tx A i) X L)" and
      ?evb = "(trecv, Recv (Rx A i) X)"
  show ?case
  proof cases
    assume "?evb \<in> set (?x # tr)"
    hence "?evb \<in> set tr" by auto
    with prems show ?case apply - apply (rule Fake.hyps(2)) by auto
  next
    assume "\<not> (?evb \<in> set (?x # tr))"
    with `?eva \<in> set (?x # tr) \<or> ?evb \<in> set (?x # tr)`
      have "?eva \<in> set (?x # tr)" by auto
    show ?case
    proof cases
      assume "?x=?eva"
      hence "X=mintr" and "Intruder I=A" and "i=j" by auto
      hence xdy: "X \<in> DM A (knowsI A tr)" using prems by auto
      show ?case
      proof cases
        assume "Key (symKey (Honest B) (Honest C)) \<in> parts {X}"
        hence ex: "\<exists> Z \<in> knowsI A tr. Key (symKey (Honest B) (Honest C)) \<in> parts {Z}"
	  using xdy `Intruder I=A` apply -
	  apply (subgoal_tac "Key (symKey (Honest B) (Honest C)) \<in> parts (DM A (knowsI A tr))")
	  apply (drule key_parts_DM_key)
	  apply (drule parts.singleton) back
	  apply auto
	  apply (subgoal_tac "{X} \<subseteq> (DM (Intruder I) (knowsI (Intruder I) tr))")
	  apply (drule parts.mono)
	  apply (erule subsetD)
	  by auto
	then obtain Z where knowsIz: "Z \<in> knowsI A tr"
	                and partsz: " Key (symKey (Honest B) (Honest C)) \<in> parts {Z}" by auto
        have ex: "(\<exists>t i. (t, Recv (Rx A i) Z) \<in> set tr) \<or> Z \<in> initState A"
          using knowsIz apply - apply (rule knowsI_A_imp_Recv_initState) by auto
        from partsz `Intruder I=A` have "Z \<notin> initState A" apply auto
          apply (drule_tac H="initState (Intruder I)" in parts.trans) apply force
	  apply (drule parts_subset_subterms[THEN subsetD])
          apply (auto simp add: symKey_notknown_other_subterms)
	  done
        with ex have "\<exists>t k. (t, Recv (Rx A k) Z) \<in> set tr" by auto
        then obtain t k where recvtr: "(t, Recv (Rx A k) Z) \<in> set tr" by auto
        with prems have  "Key (symKey (Honest B) (Honest C)) \<notin> parts {Z}"
          apply - apply (rule Fake.hyps(2)) by auto
        with partsz show ?thesis by contradiction
      qed
    next
      assume "?x\<noteq>?eva"
      with `?eva \<in> set (?x#tr)` have "?eva \<in> set tr" by auto
      with prems show ?case apply - apply (rule Fake.hyps(2)) by auto
    qed
  qed
next
  case  (Con tr tcrecv tsend E i Y L D j dCD)
  let ?x = "(tcrecv, Recv (Rx D j) Y)" and
      ?eva="(tsend, Send (Tx A i) X L)" and ?evb="(trecv, Recv (Rx A i) X)"
  show ?case
  proof cases
    assume "?eva \<in> set (?x # tr)"
    hence "?eva \<in> set tr" by auto
    thus "Key (symKey (Honest B) (Honest C)) \<notin> parts {X}" apply -
      apply (rule_tac X=X in prems(5)) by auto
  next
    assume "\<not> (?eva \<in> set (?x # tr))"
    with `?eva \<in> set (?x # tr) \<or> ?evb \<in> set (?x # tr)` have "?evb \<in> set (?x # tr)"
      by auto
    show ?case
    proof cases
      assume "?x=?evb"
      hence "Y=X" by auto
      with prems have "?x#tr \<in> sys" apply - apply (rule sys.Con) by (auto)
      with `?x=?evb` have "?evb#tr \<in> sys" by auto
      with prems
        have "\<exists>F k tcsend Lf. (tcsend, Send (Tx F k) X Lf) \<in> set (?x#tr)
              \<and> trecv >= cdist (Tx F k) (Rx A i) \<and> cdistM (Tx F k) (Rx A i) \<noteq> None \<and>
                tcsend \<le> trecv - cdist (Tx F k) (Rx A i)"
      by (intro send_before_recv) (auto)
      then obtain F k tcsend Lf
        where   "(tcsend, Send (Tx F k) X Lf) \<in> set (?x#tr)" by auto
      hence p1: "(tcsend, Send (Tx F k) X Lf) \<in> set tr" by auto
      thus ?case apply - apply (rule prems(5)) by auto
    next
      assume "?x\<noteq>?evb"
      with `?evb \<in> set (?x#tr)` have "?evb \<in> set tr" by auto
      thus ?case apply - apply (rule prems(5)) by auto
    qed
  qed
next
  case (Proto tr t' step m pEv A')
  let ?x = "(t', createEv A' pEv m)" and
      ?eva="(tsend, Send (Tx A i) X L)" and ?evb="(trecv, Recv (Rx A i) X)"
  show ?case
  proof cases -- {* Recv event already in prefix of the trace, use IH *}
    assume "?evb \<in> set (?x # tr)"
    hence "?evb \<in> set tr" by (auto simp add: createEv.simps)
    thus ?case
      apply - apply (rule Proto.hyps(2) [where X=X and B=B]) by auto
  next -- {* Send event in trace *}
    assume "\<not> (?evb \<in> set (?x # tr))"
    with `?eva \<in> set (?x # tr) \<or> ?evb \<in> set (?x # tr)` have "?eva \<in> set (?x # tr)"
      by auto
    show ?case
    proof (cases pEv)
      fix tid list assume eveq: "pEv = SendEv tid list"
      show ?case
      proof cases
        assume eqeva: "?eva=?x"
	  -- {* send event added by Proto rule, use proto\_nokeys*}
        with prems eqeva have "?x#tr \<in> sys" apply - apply (rule sys.Proto)
          by (auto)
        moreover
        with eqeva have "?eva \<in> set (?x#tr)" by auto
        ultimately show ?case using eqeva
          apply -
	  apply auto
	  apply (subgoal_tac "Key (symKey (Honest B) (Honest C)) \<in> parts (messagesProtoTr proto tr)")
	  apply (drule protoSendNoKeys)
	  apply auto prefer 2
	  apply (auto simp add: messagesProtoTr_def messagesProtoTrHonest_def MACM_def
                      split: event.split split_if dest: parts.fst_set)
	  apply (insert prems(7,8))
	  apply (case_tac pEv, auto)
	  apply (rule_tac x=A' in exI) 	apply (rule_tac x="clocktime A' t'" in exI)
	  apply (rule_tac x=step in bexI) prefer 2
	  apply force
	  apply (rule parts.mono_elem)
	  apply force
	  apply auto
	  apply (drule tuple_fst_elem)
	  apply force
	  apply (subgoal_tac "Key (symKey (Honest B) (Honest C)) \<notin> parts {M}") prefer 2
	  apply (rule Proto.hyps(2))
	  apply force
	  apply force
	  done
      next
        assume "?eva\<noteq>?x"
        with `?eva \<in> set (?x # tr)` have "?eva \<in> set tr" by auto
        with Proto.hyps(2) [where X=X and B=B and L=L] show ?case by auto
      qed
    next
      assume "pEv = ClaimEv"
      hence "?x\<noteq>?eva" by auto
      with `?eva \<in> set (?x#tr)` have "?eva \<in> set tr" by auto
      with Proto.hyps(2) [where X=X and B=B and L=L] show ?case by auto
    qed
  qed
qed

lemma (in PROTOCOL_PKSIG_NOKEYS) key_not_known:
  assumes sys_proto: "tr \<in> sys" and neq: "A \<noteq> Honest B"
  shows "Key (priSK (Honest B)) \<notin> parts (knowsI A tr)" using sys_proto
proof auto
  assume "Key (priSK (Honest B)) \<in> parts (knowsI A tr)"
  hence "\<exists> X \<in> knowsI A tr. Key (priSK (Honest B)) \<in> parts {X}"
    by (rule parts.singleton)
  then obtain X where "X \<in> knowsI A tr"
                  and partsx: "Key (priSK (Honest B)) \<in> parts {X}" by auto
  hence ex: "(\<exists>t i. (t, Recv (Rx A i) X) \<in> set tr) \<or> X \<in> initState A"
          apply - apply (rule knowsI_A_imp_Recv_initState) by auto
  show "False" proof cases
    assume "\<exists>t i. (t, Recv (Rx A i) X) \<in> set tr"
    then obtain t i where "(t, Recv (Rx A i) X) \<in> set tr" by auto
    with sys_proto have "Key (priSK (Honest B)) \<notin> parts {X}" apply -
      apply (rule keys_not_send_received) by (auto)
    with partsx show False by auto
  next
    assume "\<not>(\<exists>t i. (t, Recv (Rx A i) X) \<in> set tr)"
    with ex have  kinit: "X \<in> initState A" by auto
    from neq partsx have "X \<notin> initState A"
      apply auto
      apply (drule_tac H="initState A" in parts.trans) apply simp
      apply force
      apply (drule parts_subset_subterms[THEN subsetD])
      apply (auto dest: priSK_notknown_other_subterms)
      done
    with kinit show False by contradiction
  qed
qed


lemma (in PROTOCOL_SYMKEYS_NOKEYS) key_not_known:
  assumes sys_proto: "tr \<in> sys" and neq: "A \<notin> {Honest B, Honest C}"
  shows "Key (symKey (Honest B) (Honest C)) \<notin> parts (knowsI A tr)" using sys_proto
proof auto
  assume "Key (symKey (Honest B) (Honest C)) \<in> parts (knowsI A tr)"
  hence "\<exists> X \<in> knowsI A tr. Key (symKey (Honest B) (Honest C)) \<in> parts {X}"
    by (rule parts.singleton)
  then obtain X where "X \<in> knowsI A tr"
                  and partsx: "Key (symKey (Honest B) (Honest C)) \<in> parts {X}" by auto
  hence ex: "(\<exists>t i. (t, Recv (Rx A i) X) \<in> set tr) \<or> X \<in> initState A"
          apply - apply (rule knowsI_A_imp_Recv_initState) by auto
  show "False" proof cases
    assume "\<exists>t i. (t, Recv (Rx A i) X) \<in> set tr"
    then obtain t i where "(t, Recv (Rx A i) X) \<in> set tr" by auto
    with sys_proto have "Key (symKey (Honest B) (Honest C)) \<notin> parts {X}" apply -
      apply (rule keys_not_send_received) by (auto)
    with partsx show False by auto
  next
    assume "\<not>(\<exists>t i. (t, Recv (Rx A i) X) \<in> set tr)"
    with ex have  kinit: "X \<in> initState A" by auto
    from neq partsx have "X \<notin> initState A"
      apply auto
      apply (drule_tac H="initState A" in parts.trans) apply force
      apply (drule parts_subset_subterms[THEN subsetD])
      apply (auto dest: symKey_notknown_other_subterms)
      done
    with kinit show False by contradiction
  qed
qed


lemma (in PROTOCOL_PKSIG_NOKEYS) sig_generate_sig_received:
 assumes sys_proto: "tr \<in> sys" and syn: "m \<in> DM B (knowsI B tr)"
         and sig: "Crypt (priSK (Honest A)) msig \<in> subterms {m}"
         and neq: "B \<noteq> Honest A"
 shows "\<exists>trs X i. (trs, Recv (Rx B i) X) \<in> set tr
                  \<and> Crypt (priSK (Honest A)) msig \<in> subterms {X}"
  using sys_proto syn sig neq
proof -
  from syn sig
    have sig_or_key: "Crypt (priSK (Honest A)) msig \<in> subterms (knowsI B tr)
                      \<or> Key (priSK (Honest A)) \<in> parts (knowsI B tr)"
      using prems
      apply -
      apply (subgoal_tac "Crypt (priSK (Honest A)) msig \<in> subterms (DM B (knowsI B tr))") prefer 2
      apply (erule rev_subsetD)
      apply (rule subterms.mono)
      apply force
      apply auto
      apply (drule sig_subterms_DM_sig_or_key)
      apply auto
      done
 hence "Crypt (priSK (Honest A)) msig \<in> subterms (knowsI B tr)"
   using key_not_known prems by auto
 hence "\<exists> X \<in> knowsI B tr. Crypt (priSK (Honest A)) msig \<in> subterms {X}"
   by (auto intro: subterms.singleton)
 then obtain X where knowsX: "X \<in> knowsI B tr"
                 and sigX:   "Crypt (priSK (Honest A)) msig \<in> subterms {X}"
   by auto
 hence "(\<exists> t i. (t, Recv (Rx B i) X) \<in> set tr)"
   apply -
   apply (drule knowsI_A_imp_Recv_initState)
   apply auto
   apply (drule_tac H="initState B" in subterms.trans)
   apply (insert priSK_not_used, auto)
   done
 thus ?thesis using sigX by auto
qed

lemma (in PROTOCOL_SYMKEYS_NOKEYS) mac_generate_mac_received:
 assumes sys_proto: "tr \<in> sys" and syn: "m \<in> DM B (knowsI B tr)"
         and sig: "Hash (MPair (Key (symKey (Honest C) (Honest D))) mmac) \<in> subterms {m}"
         and neq: "B \<notin> {Honest C, Honest D}"
 shows "\<exists>trs X i. (trs, Recv (Rx B i) X) \<in> set tr
                  \<and> Hash (MPair (Key (symKey (Honest C) (Honest D))) mmac) \<in> subterms {X}"
  using sys_proto syn sig neq
proof -
  let ?key = "(Key (symKey (Honest C) (Honest D)))"
  let ?mac = "Hash (MPair ?key mmac)"
  from syn sig
    have sig_or_key: "?mac \<in> subterms (knowsI B tr)
                      \<or> ?key \<in> parts (knowsI B tr)"
      using prems
      apply (subgoal_tac "?mac \<in> subterms (DM B (knowsI B tr))")
      apply (drule mac_subterms_DM_mac_or_key)
      apply auto
      apply (erule rev_subsetD)
      apply (rule subterms.mono, auto)
      done
 hence "?mac \<in> subterms (knowsI B tr)"
   using key_not_known prems by auto
 hence "\<exists> X \<in> knowsI B tr. ?mac \<in> subterms {X}"
   by (auto intro: subterms.singleton)
 then obtain X where knowsX: "X \<in> knowsI B tr"
                 and sigX:   "?mac \<in> subterms {X}"
   by auto
 hence "(\<exists> t i. (t, Recv (Rx B i) X) \<in> set tr)"
   apply -
   apply (drule knowsI_A_imp_Recv_initState)
   apply auto
   apply (drule_tac H="initState B" in subterms.trans)
   apply force
   apply (insert symKey_not_used_MAC)
   apply auto
   done
 thus ?thesis using sigX by auto
qed
  
locale PROTOCOL_NONONCE = INITSTATE_NONONCE + PROTOCOL

lemma (in PROTOCOL_NONONCE) nonce_orig_not_before:
 assumes A: "(ta, Send A X La) \<in> set tr" and B: "Nonce C NC \<in> subterms {X}" and
         C: "Nonce C NC \<notin> used (beforeEvent (tb, Send B Y Lb) tr)"
 shows "(ta, Send A X La) \<notin> set (beforeEvent (tb, Send B Y Lb) tr)" using A B C
proof cases
  assume "(ta, Send A X La) \<in> set (beforeEvent (tb, Send B Y Lb) tr)"
  with B have "Nonce C NC \<in> used (beforeEvent (tb, Send B Y Lb) tr)" apply -
    apply (rule Send_imp_parts_used) 
    by auto
  thus ?thesis using C by contradiction
next
  assume "(ta, Send A X La) \<notin> set (beforeEvent (tb, Send B Y Lb) tr)" 
  thus ?thesis .
qed

lemma (in PROTOCOL_NONONCE) nonce_send_owner_first:
  assumes a: "tr \<in> sys" and b: "(tb,Send (Tx B i) mb Lb) \<in> set tr" and
          c: "Nonce A NA \<in> subterms {mb}" and d: "A \<noteq> B"
  shows "\<exists>j ta ma La. (ta,Send (Tx A j) ma La) \<in> set tr \<and> Nonce A NA \<in> subterms {ma}"
  using a b c d
proof (induct tr arbitrary: A B tb mb i NA Lb rule: sys.induct)
-- {* @{text "trace equal to: @{term (tc + tab, Recv D mc)#tr"} *}
  case Nil thus ?case by auto
next
  case  (Con tr tcrev tsend C k Y Ly D l dCD)
  hence oin: "(tb, Send (Tx B i) mb Lb) \<in> set tr" by auto
  with Con.hyps prems show ?case by (auto)
next
  case (Fake tr tsend mspy I k)
  let ?x = "(tsend, Send (Tx (Intruder I) k) mspy [])" and
      ?evb = "(tb, Send (Tx B i) mb Lb)"
  show ?case
  proof cases
    assume "?x = ?evb"
    hence "mspy=mb" and "Intruder I=B" by auto
    with prems
      have "\<exists>t i Y. (t, Recv (Rx B i) Y) \<in> set tr \<and> Nonce A NA \<in> subterms {Y}"
      apply -
      apply (rule othernonce_gen_received) by (auto)
    then obtain t k Y where "(t, Recv (Rx B k) Y) \<in> set tr" and
                            "Nonce A NA \<in> subterms {Y}"  by auto
    with prems have "\<exists>C u tcsend Lc. (tcsend, Send (Tx C u) Y Lc) \<in> set tr \<and>
        t >= cdist (Tx C u) (Rx B k) \<and> cdistM (Tx C u) (Rx B k) \<noteq> None \<and>
        tcsend \<le> t - cdist (Tx C u) (Rx B k)"
	by (intro send_before_recv) (auto)
    then obtain C u tcsend Lc
	where p1: "(tcsend, Send (Tx C u) Y Lc) \<in> set tr" by auto
    show ?case
    proof cases
      assume "A=C"
      with p1 `Nonce A NA \<in> subterms {Y}` show ?thesis by auto
    next
      assume "A\<noteq>C"
      with prems `Nonce A NA \<in> subterms {Y}`
        have "\<exists>j ta ma La. (ta, Send (Tx A j) ma La) \<in> set tr
                        \<and> Nonce A NA \<in> subterms {ma}" apply -
	  by auto
      then obtain j ta ma La where a: "(ta, Send (Tx A j) ma La) \<in> set tr" and
                                   b: "Nonce A NA \<in> subterms {ma}" by auto
      then have "(ta, Send (Tx A j) ma La) \<in> set (?x#tr)" by auto
      with prems show ?thesis by auto
    qed
  next
    assume "?x \<noteq> ?evb"
    with prems have "?evb \<in> set tr" by auto
    with prems have "\<exists>j ta ma La. (ta, Send (Tx A j) ma La) \<in> set tr
                               \<and> Nonce A NA \<in> subterms {ma}" by auto
    then obtain j ta ma La where a: "(ta, Send (Tx A j) ma La) \<in> set tr" and
                                 b: "Nonce A NA \<in> subterms {ma}" by auto
    then have "(ta, Send (Tx A j) ma La) \<in> set (?x#tr)" by auto
    with prems show ?thesis by auto
  qed
next
  case (Proto tr tsend step m pEv C)
  let ?x = "(tsend, createEv C pEv m)" and
      ?evb = "(tb, Send (Tx B i) mb Lb)"
  show ?case
  proof cases
    assume "?x = ?evb"
    hence "m=mb" and "Honest C=B"
      apply - by (case_tac "pEv", auto simp add: createEv.simps)+
    with prems
      have "\<exists>t i Y. (t, Recv (Rx B i) Y) \<in> set tr \<and> Nonce A NA \<in> subterms {Y}"
      apply - apply (rule othernonce_gen_received) by (auto)
    then obtain t k Y where "(t, Recv (Rx B k) Y) \<in> set tr" and
                            "Nonce A NA \<in> subterms {Y}"  by auto
    with prems have "\<exists>D u tcsend Lc. (tcsend, Send (Tx D u) Y Lc) \<in> set tr \<and>
        t >= cdist (Tx D u) (Rx B k) \<and> cdistM (Tx D u) (Rx B k) \<noteq> None \<and>
        tcsend \<le> t - cdist (Tx D u) (Rx B k)"
	by (intro send_before_recv) (auto)
    then obtain D u tcsend Lc
	where p1: "(tcsend, Send (Tx D u) Y Lc) \<in> set tr" by auto
    show ?case
    proof cases
      assume "A=D"
      with p1 `Nonce A NA \<in> subterms {Y}` show ?thesis by auto
    next
      assume "A\<noteq>D"
      with prems `Nonce A NA \<in> subterms {Y}`
        have "\<exists>j ta ma La. (ta, Send (Tx A j) ma La) \<in> set tr
                           \<and> Nonce A NA \<in> subterms {ma}" by auto
      then obtain j ta ma La where a: "(ta, Send (Tx A j) ma La) \<in> set tr" and
                                   b: "Nonce A NA \<in> subterms {ma}" by auto
      then have "(ta, Send (Tx A j) ma La) \<in> set (?x#tr)" by auto
      with prems show ?thesis by auto
    qed
  next
    assume "?x \<noteq> ?evb"
    with prems have "?evb \<in> set tr" by auto
    with prems have "\<exists>j ta ma La. (ta, Send (Tx A j) ma La) \<in> set tr
                                  \<and> Nonce A NA \<in> subterms {ma}" by auto
    then obtain j ta ma La where a: "(ta, Send (Tx A j) ma La) \<in> set tr" and
                                 b: "Nonce A NA \<in> subterms {ma}" by auto
    then have "(ta, Send (Tx A j) ma La) \<in> set (?x#tr)" by auto
    with prems show ?thesis by auto
  qed
qed

lemma (in PROTOCOL_NONONCE) Used_imp_subterms_Send_creator:
  assumes a: "Nonce A NA \<in> used tr" and b: "tr \<in> sys"
  shows "\<exists>i t X L. (t, Send (Tx A i) X L) \<in> set tr \<and> Nonce A NA \<in> subterms {X}"
  using a b
proof -
  from prems
    have "\<exists>t B i X L. (t, Send (Tx B i) X L) \<in> set tr \<and> Nonce A NA \<in> subterms {X}"
    apply - apply (rule Used_imp_subterm_Send) by auto
  then obtain t B i X L
    where c: "(t, Send (Tx B i) X L) \<in> set tr" and d: "Nonce A NA \<in> subterms {X}"
    apply auto
    done
  show ?thesis
  proof cases
    assume "A=B"
    with c d show ?thesis by auto
  next
    assume "A\<noteq>B"
    with prems show ?thesis apply - apply (rule nonce_send_owner_first) by auto
  qed
qed

lemma (in PROTOCOL_NONONCE) nonce_used_view:
  "\<lbrakk> tr \<in> sys; Nonce (Honest A) NA \<in> used tr\<rbrakk>
  \<Longrightarrow> Nonce (Honest A) NA \<in> used (view A tr)"
  apply (drule Used_imp_subterms_Send_creator)
  apply (force, elim exE conjE)
  apply (drule view_elem_at_ex)
  apply force
  apply (elim exE)
  apply (rule_tac X=X and RA="Tx (Honest A) i" and t="t'" in subterms_set_used)
  apply auto
done

text {* now we get to the first important property concerning the
        reply to messages including fresh nonces. *}

lemma (in PROTOCOL_NONONCE) fresh_nonce_earliest_send:
  assumes sys_proto: "tr \<in> sys" and aneqb: "A\<noteq>B" and
          nafresh:  "Nonce A NA \<notin> used (beforeEvent (ta, Send (Tx A i) ma La) tr)" and
          na_in_ma: "Nonce A NA \<in> subterms {ma}" and
          na_in_mb: "Nonce A NA \<in> subterms {mb}" and
          eva: "(ta, Send (Tx A i) ma La) \<in> set tr" and evb: "(tb, Send (Tx B j) mb Lb) \<in> set tr"
  shows "tb - ta >= cdistl A B"
    using sys_proto aneqb nafresh na_in_ma na_in_mb eva evb
proof (induct tr arbitrary: A B ta tb ma mb La Lb i j NA rule: sys.induct) 
  case Nil thus ?case by auto

-- {* @{text "trace equal to: @{term (tc + tab, Recv D mc)#tr"} *}
next
  case  (Con tr tsend C k Y Ld D l dCD tcrecv)
  hence oin: "(ta, Send (Tx A i) ma La) \<in> set tr" and 
        rin: "(tb, Send (Tx B j) mb Lb) \<in> set tr" and
        nafresh: "Nonce A NA \<notin> used (beforeEvent (ta, Send (Tx A i) ma La) tr)" by auto
  with Con.hyps prems show ?case by (auto)

-- {* @{text "(tsend, Send Intruder I mspy)#tr"} *}
next
  case (Fake tr  tsend mspy I k)
  let ?x = "(tsend, Send (Tx (Intruder I) k) mspy [])" and
      ?eva = "(ta, Send (Tx A i) ma La)" and ?evb = "(tb, Send (Tx B j) mb Lb)"

  note caserule =  set_two_elem_cases [where eva="?eva" and evb="?evb"
                                       and tr="tr" and x="?x"]
  from prems show ?case
  proof (cases rule: caserule, simp, simp)
    case 3
    note `?eva \<in> set tr` and `?evb \<in> set tr`
    moreover
    from Fake.prems have "Nonce A NA \<notin> used (beforeEvent ?eva tr)"
    proof cases
      assume evaeq: "?eva = ?x \<and> ?eva \<notin> set tr"
      with Fake.prems have "Nonce A NA \<notin> used tr" by (clarsimp)
      with evaeq show ?thesis by (intro used_beforeEvent)
    next
      assume "\<not> (?eva = ?x \<and> ?eva \<notin> set tr)"
      thus ?thesis using prems by auto
    qed
    ultimately show ?thesis using Fake.hyps Fake.prems by auto
  next
    case 4
    note `?eva \<in> set tr` and beq=`?evb = ?x` and `?eva \<noteq> ?x`
    with prems have "Nonce A NA \<in> subterms {mspy}" by auto
    with beq have "Intruder I = B" and "tsend = tb" and "j=k" by auto
    from `Nonce A NA \<in> subterms {ma}` `?eva \<in> set tr`
      have "Nonce A NA \<in> used tr" apply - apply (rule Send_imp_parts_used) by auto
    with prems have "\<exists>trs l X. (trs, Recv (Rx (Intruder I) l) X) \<in> set tr \<and>
                               Nonce A NA \<in> subterms {X}"
      apply - apply (rule_tac X="mspy" in othernonce_gen_received) by (auto)
    with beq obtain X l trs
      where recv: "(trs, Recv (Rx B l) X) \<in> set tr" and
            naX: "Nonce A NA \<in> subterms {X}"
      by auto
    with prems
    have "\<exists>C u tcsend Lc. (tcsend, Send (Tx C u) X Lc) \<in> set tr \<and>
        trs >= cdist (Tx C u) (Rx B l) \<and> cdistM (Tx C u) (Rx B l) \<noteq> None \<and>
        tcsend \<le> trs - cdist (Tx C u) (Rx B l)"
	by (intro send_before_recv) (auto)
    then obtain C u tcsend Lc
	where p1: "(tcsend, Send (Tx C u) X Lc) \<in> set tr" and
        p2: "trs >= cdist (Tx C u) (Rx B l)" and
        p3: "cdistM (Tx C u) (Rx B l) \<noteq> None" and
        p4: "tcsend \<le> trs - cdist (Tx C u) (Rx B l)" by auto
    let ?evc="(tcsend, Send (Tx C u) X Lc)"
    show ?thesis
    proof cases
      assume "A=C"
      with prems `?eva\<noteq>?x` have nafresh: "Nonce A NA \<notin> used (beforeEvent ?eva tr)"
        by auto
      with p1 naX `A=C` have "?evc \<notin> set (beforeEvent ?eva tr)"
        apply - apply (rule nonce_orig_not_before) by auto
      with prems p1 `A=C` have p5: "tcsend \<ge> ta"
        apply - apply (rule not_beforeEvent_later) by auto
      from p3 p4 have "cdist (Tx C u) (Rx B l) \<le> trs - tcsend" by auto
      with p5 also have p6: "cdist (Tx C u) (Rx B l) \<le> trs - ta" by auto
      with `maxtime tr <= tsend` and `tsend=tb` and `(trs, Recv (Rx B l) X) \<in> set tr`
      have tbgeq: "trs <= tb" by (auto intro: maxtime_geq_elem)
      with tbgeq p6 have p7: "cdist (Tx C u) (Rx B l) \<le> tb - ta" by auto
      with p3 have "cdistl C B \<le> cdist (Tx C u) (Rx B l)" apply -
	apply (unfold cdist_def,rule noflt_some)
	by auto
      with p7 `A=C` show ?thesis by auto
    next -- {* use induction hypothesis with Send A and Send C to bound trecv *}
      assume "A\<noteq>C"
      with p1 and prems have "cdistl A C \<le> tcsend - ta"
        apply - apply (rule Fake.hyps) by auto
      with p4 have dsmaller: "cdistl A C \<le> trs - cdist (Tx C u) (Rx B l) - ta" by auto
      with `maxtime tr <= tsend` and `tsend=tb` and `(trs, Recv (Rx B l) X) \<in> set tr`
      have tbgeq: "trs <= tb" by (auto intro: maxtime_geq_elem)
      from p3 have "cdist (Tx C u) (Rx B l) \<ge> cdistl C B" apply -
	by (unfold cdist_def, rule noflt_some)
      from dsmaller p1 have a1: "cdistl A C \<le> trs - ta - cdist (Tx C u) (Rx B l)"
	by (arith)
      also with `cdist (Tx C u) (Rx B l) \<ge> cdistl C B`
        have "\<dots> \<le>  trs - ta - cdistl C B" by arith
      finally have q: "cdistl A C + cdistl C B \<le> trs - ta" by auto
      hence "cdistl A C + cdistl C B \<le> trs - ta" by auto
      with cdistl_triangle have "cdistl A B \<le> cdistl A C + cdistl C B" by auto
      hence "cdistl A B \<le> cdistl A C + cdistl C B" by auto
      also with q have "\<dots> \<le> trs - ta" by auto
      also with tbgeq have "\<dots> \<le> tb - ta" by auto
      finally show ?thesis by auto
    qed
  next
    case 5
    note `?evb \<in> set tr` and `?eva = ?x` and `?evb \<noteq> ?x`
    hence "A=Intruder I" by auto
    with `Nonce A NA \<in> subterms {mb}` have used: "Nonce A NA \<in> used tr" using prems
      apply - apply (rule Send_imp_parts_used) by auto
    show ?thesis proof cases
      assume "?eva \<in> set tr"
      hence "Nonce A NA \<notin> used (beforeEvent ?eva tr)" using prems by auto
      thus ?thesis using `?evb \<in> set tr` `A\<noteq>B` apply -
	apply (rule Fake.hyps(2))
	apply assumption+ prefer 4
	apply assumption
	apply (auto intro: prems)
	done
    next
      assume "?eva \<notin> set tr"
      with `?eva = ?x` used have "Nonce A NA \<in> used (beforeEvent ?eva (?x#tr))" by auto
      thus ?thesis using `Nonce A NA \<notin> used (beforeEvent ?eva (?x#tr))`
	by contradiction
    qed
  next
    case 6
    note `?eva = ?x` and `?evb = ?x` and `A\<noteq>B`
    hence "A = B" by auto
    thus ?thesis using `A\<noteq>B` by contradiction
  qed
next
  case (Proto tr tsend step m pEv C)
  let ?x = "(tsend, createEv C pEv m)" and
      ?eva = "(ta, Send (Tx A i) ma La)" and ?evb = "(tb, Send (Tx B j) mb Lb)"
  note caserule =  set_two_elem_cases [where eva="?eva" and evb="?evb"
                                       and tr="tr" and x="?x"]
  from prems show ?case
  proof (cases rule: caserule, simp, simp)
    case 3
    note `?eva \<in> set tr` and `?evb \<in> set tr`
    show ?thesis
    proof cases
      assume evaeq: "?eva = ?x \<and> ?eva \<notin> set tr"
      with prems have "Nonce A NA \<notin> used tr" by (clarsimp)
      hence "Nonce A NA \<notin> used (beforeEvent ?eva tr)" by (intro used_beforeEvent)
      with prems `?eva \<in> set tr` and `?evb \<in> set tr` show ?thesis apply -
        apply (rule_tac ma=ma and mb=mb in Proto.hyps(2)) .
    next
      assume "\<not> (?eva = ?x \<and> ?eva \<notin> set tr)"
      thus ?thesis using prems by auto
    qed
  next
    case 6
    note `?eva = ?x` and `?evb = ?x`
    hence "A = B" apply auto apply (case_tac "pEv") by auto
    thus ?thesis using `A\<noteq>B` by contradiction
  next
    case 5
    note `?evb \<in> set tr` and `?eva = ?x` and `?evb \<noteq> ?x`
    with `Nonce A NA \<in> subterms {mb}` have used: "Nonce A NA \<in> used tr"
      apply - apply (rule Send_imp_parts_used) by auto
    show ?thesis proof cases
      assume "?eva \<in> set tr"
      hence "Nonce A NA \<notin> used (beforeEvent ?eva tr)" using `?eva = ?x` prems by auto
      thus ?thesis using `?evb \<in> set tr` prems apply -
	apply (rule Proto.hyps(2)) prefer 6
	apply assumption+
	done
    next
      assume "?eva \<notin> set tr"
      with `?eva = ?x` used have "Nonce A NA \<in> used (beforeEvent ?eva (?x#tr))" by auto
      thus ?thesis using `Nonce A NA \<notin> used (beforeEvent ?eva (?x#tr))`
      by contradiction
    qed
  next
    case 4
    note `?eva \<in> set tr` and beq=`?evb = ?x` and `?eva \<noteq> ?x`
    with prems have "Nonce A NA \<in> subterms {m}"
      apply (case_tac "pEv")
      apply auto done
    from `?evb = ?x` have "\<exists> k L. pEv = SendEv k L" apply -
      apply (case_tac "pEv") by auto
    then obtain k Le where  pev: "pEv = SendEv k Le" by auto
    with beq have "Honest C=B" and "tsend = tb" and "j=k" by (auto)
      
    -- {* either the nonce has been already sent or received  *}
    from `Nonce A NA \<in> subterms {ma}` `?eva \<in> set tr`
      have "Nonce A NA \<in> used tr" apply - apply (rule Send_imp_parts_used) by auto
    with prems have "\<exists>trs l X. (trs, Recv (Rx B l) X) \<in> set tr \<and>
      Nonce A NA \<in> subterms {X}"
      apply - apply (rule_tac X="m" in othernonce_gen_received) apply force
      apply force by (auto)
    with beq obtain X l trs
      where recv: "(trs, Recv (Rx B l) X) \<in> set tr" and
            naX: "Nonce A NA \<in> subterms {X}"
	by auto
    with prems
      have "\<exists>C u tcsend Lc. (tcsend, Send (Tx C u) X Lc) \<in> set tr \<and>
        trs >= cdist (Tx C u) (Rx B l) \<and> cdistM (Tx C u) (Rx B l) \<noteq> None \<and>
        tcsend \<le> trs - cdist (Tx C u) (Rx B l)"
	by (intro send_before_recv) (auto)
    then obtain D u tcsend Lc
	where p1: "(tcsend, Send (Tx D u) X Lc) \<in> set tr" and
        p2: "trs >= cdist (Tx D u) (Rx B l)" and
        p3: "cdistM (Tx D u) (Rx B l) \<noteq> None" and
        p4: "tcsend \<le> trs - cdist (Tx D u) (Rx B l)" by auto
    let ?evc="(tcsend, Send (Tx D u) X Lc)"
    show ?thesis
    proof cases
      assume "A=D"
      with prems `?eva\<noteq>?x` have nafresh: "Nonce A NA \<notin> used (beforeEvent ?eva tr)"
        by auto
      with p1 naX `A=D` have "?evc \<notin> set (beforeEvent ?eva tr)"
        apply - apply (rule nonce_orig_not_before) by auto
      with prems p1 `A=D` have p5: "tcsend \<ge> ta"
        apply - apply (rule not_beforeEvent_later) by auto
      from p3 p4 have "cdist (Tx D u) (Rx B l) \<le> trs - tcsend" by auto
      with p5 also have p6: "cdist (Tx D u) (Rx B l) \<le> trs - ta" by auto
      with `maxtime tr <= tsend` and `tsend=tb`
           and `(trs, Recv (Rx B l) X) \<in> set tr`
      have tbgeq: "trs <= tb" by (auto intro: maxtime_geq_elem)
      with tbgeq p6 have p7: "cdist (Tx D u) (Rx B l) \<le> tb - ta" by auto
      with p3 have "cdistl D B \<le> cdist (Tx D u) (Rx B l)" apply -
	by (unfold cdist_def, rule noflt_some)
      with p7 `A=D` show ?thesis by auto
    next -- {* use induction hypothesis with Send A and Send C to bound trecv *}
      assume "A\<noteq>D"
      hence "cdistl A D \<le> tcsend - ta" using p1 and prems apply - by auto
      with p4 have dsmaller: "cdistl A D \<le> trs - cdist (Tx D u) (Rx B l) - ta" by auto
      with `maxtime tr <= tsend` and `tsend=tb`
           and `(trs, Recv (Rx B l) X) \<in> set tr`
      have tbgeq: "trs <= tb" by (auto intro: maxtime_geq_elem)
      from p3 have "cdist (Tx D u) (Rx B l) \<ge> cdistl D B" apply -
	by (unfold cdist_def, rule noflt_some)
      from dsmaller p1 have a1: "cdistl A D \<le> trs - ta - cdist (Tx D u) (Rx B l)"
	by (arith)
      also with `cdist (Tx D u) (Rx B l) \<ge> cdistl D B`
        have "\<dots> \<le>  trs - ta - cdistl D B" by arith
      finally have q: "cdistl A D + cdistl D B \<le> trs - ta" by auto
      hence "cdistl A D + cdistl D B \<le> trs - ta" by auto
      with cdistl_triangle have "cdistl A B \<le> cdistl A D + cdistl D B" by auto
      hence "cdistl A B \<le> cdistl A D + cdistl D B" by auto
      also with q have "\<dots> \<le> trs - ta" by auto
      also with tbgeq have "\<dots> \<le> tb - ta" by auto
      finally show ?thesis by auto
    qed
  qed
qed

lemma (in PROTOCOL_PKSIG_NOKEYS) sig_earliest_send:
  assumes sys_proto: "tr \<in> sys" and
          mbsig: "(Crypt (priSK (Honest A)) msig) \<in> subterms {mb}" and
          mbsent: "(tb, Send (Tx B j) mb Lb) \<in> set tr" and
          aneqb: "Honest A\<noteq>B"
  shows "\<exists>ta ma i La. tb - ta >= cdistl (Honest A) B \<and>
         (ta, Send (Tx (Honest A) i) ma La) \<in> set tr
         \<and> (Crypt (priSK (Honest A)) msig) \<in> subterms {ma}"
  using sys_proto mbsig mbsent aneqb
proof (induct tr arbitrary: A B j tb mb msig Lb rule: sys.induct)
  case Nil thus ?case by auto

-- {* @{text "trace equal to: @{term (tc + tab, Recv D mc)#tr"} *}
next
  case (Con tr tc C k mc Ld D l dCD trecv)
  hence "(tb, Send (Tx B j) mb Lb) \<in> set tr" by auto
  with Con.hyps prems show ?case by (auto)
-- {* @{text "(tsend, Send Intruder I mspy)#tr"} *}
next
  case (Fake tr tsend mintr I k)
  let ?x = "(tsend, Send (Tx (Intruder I) k) mintr [])" and
      ?evreply = "(tb, Send (Tx B j) mb Lb)"

  show ?case
  proof cases
    assume xeq: "?x = ?evreply"
    with xeq have "B=Intruder I" and "mb=mintr" and "tsend=tb" by auto
    -- {* Sig in synth (Nonce I`.. u analz (knowsI I tr) => Recv X mit message
          => Send C mit message => IH*}
    with prems have "\<exists>trecv X i. (trecv, Recv (Rx B i) X) \<in> set tr
                     \<and> Crypt (priSK (Honest A)) msig \<in> subterms {X}"
      apply - apply (rule sig_generate_sig_received) by (auto)
    then obtain X trecv l
      where       "(trecv, Recv (Rx B l) X) \<in> set tr" and
            sigX: "Crypt (priSK (Honest A)) msig \<in> subterms {X}" by auto
    with prems
      have "\<exists>C u tcsend Lc. (tcsend, Send (Tx C u) X Lc) \<in> set tr
            \<and> trecv >= cdist (Tx C u) (Rx B l) \<and> cdistM (Tx C u) (Rx B l) \<noteq> None
            \<and> tcsend \<le> trecv - cdist (Tx C u) (Rx B l)"
      by (intro send_before_recv) (auto)
    then obtain C tcsend u Lc
      where p1: "(tcsend, Send (Tx C u) X Lc) \<in> set tr" and
            p2: "trecv >= cdist (Tx C u) (Rx B l)" and
            p3: "cdistM (Tx C u) (Rx B l) \<noteq> None" and
            p4: "tcsend \<le> trecv - cdist (Tx C u) (Rx B l)" by auto
    let ?evc="(tcsend, Send (Tx C u) X Lc)"
    show ?thesis
    proof cases
      assume "Honest A=C"
      with p4 have p5: "cdist (Tx C u) (Rx B l) \<le> trecv - tcsend" by auto
      with `maxtime tr <= tsend` and `tsend=tb` and `(trecv, Recv (Rx B l) X) \<in> set tr`
      have tbgeq: "trecv <= tb" by (auto intro: maxtime_geq_elem)
      with p5 have p6:"cdist (Tx C u) (Rx B l) \<le> tb - tcsend" by auto
      with p3 have "cdistl C B \<le> cdist (Tx C u) (Rx B l)" apply -
	by (unfold cdist_def, rule noflt_some)
      with p6 `Honest A=C` have "cdistl (Honest A) B \<le> tb -tcsend" by auto
      with p1 sigX `Honest A=C`
        show ?thesis apply (rule_tac x=tcsend in exI, rule_tac x=X in exI)
        apply (rule_tac x=u in exI)
        by (auto)
    next
      assume "Honest A\<noteq>C"
      with prems have "\<exists>ta ma i La. cdistl (Honest A) C \<le> tcsend - ta \<and>
         (ta, Send (Tx (Honest A) i) ma La) \<in> set tr \<and>
         Crypt (priSK (Honest A)) msig \<in> subterms {ma}" by auto
      then obtain ta ma i La where cdistl: "cdistl (Honest A) C \<le> tcsend - ta"
                             and intr: "(ta, Send (Tx (Honest A) i) ma La) \<in> set tr"
                             and sig: "Crypt (priSK (Honest A)) msig \<in> subterms {ma}"
	                     by auto
      with p4 have p5: "cdistl (Honest A) C \<le>  trecv - cdist (Tx C u) (Rx B l) - ta" by auto
      show ?thesis
      proof (rule_tac x=ta in exI, rule_tac x=ma in exI,
	     rule_tac x=i in exI, rule_tac x=La in exI, intro conjI)
        from `maxtime tr <= tsend` and `tsend=tb` and
	     `(trecv, Recv (Rx B l) X) \<in> set tr`
          have tr_tb: "trecv <= tb" by (auto intro: maxtime_geq_elem)
        from p3 noflt_some have "cdistl C B \<le> the (cdistM (Tx C u) (Rx B l))" by auto
        hence "cdistl C B \<le> cdist (Tx C u) (Rx B l)" by (auto simp add: cdist_def)
        with cdistl p5 have "cdistl (Honest A) C  \<le> trecv - ta - cdistl C B" by auto
        hence a1: "cdistl (Honest A) C + cdistl C B \<le> trecv - ta" by auto
        from cdistl_triangle
	  have "cdistl (Honest A) B \<le>  cdistl (Honest A) C + cdistl C B" by auto
        then have "cdistl (Honest A) B \<le> cdistl (Honest A) C + cdistl C B" by auto
        also with a1 have "\<dots> \<le> trecv - ta" by auto
        also with tr_tb have "\<dots> \<le> tb - ta" by auto
        finally show "cdistl (Honest A) B \<le> tb - ta" by auto
      next
        from prems show "(ta, Send (Tx (Honest A) i) ma La) \<in> set (?x # tr)" by auto
      next
        from prems show "Crypt (priSK (Honest A)) msig \<in> subterms {ma}" by auto
      qed
    qed
  next
    assume "?x \<noteq> ?evreply"
    with prems have "?evreply \<in> set tr" by auto
    with prems have "\<exists>ta ma i La. cdistl (Honest A) B \<le> tb - ta \<and>
       (ta, Send (Tx (Honest A) i) ma La) \<in> set tr \<and>
       Crypt (priSK (Honest A)) msig \<in> subterms {ma}" by auto
    thus ?case by auto
  qed
next
  case (Proto tr tsend step m pEv C)
  let ?x = "(tsend, createEv C pEv m)" and
      ?evreply = "(tb, Send (Tx B j) mb Lb)"

  show ?case
  proof cases
    assume xeq: "?x = ?evreply"
    with xeq have "B=Honest C" and "mb=m" and "tsend=tb" apply auto
      apply (case_tac "pEv") apply auto apply (case_tac "pEv")
      by auto
    -- {* Sig in synth (Nonce I`.. u analz (knowsI I tr) => Recv X mit message
          => Send C mit message => IH*}
    with prems have "\<exists>trecv X i. (trecv, Recv (Rx B i) X) \<in> set tr
                     \<and> Crypt (priSK (Honest A)) msig \<in> subterms {X}"
      apply - apply (rule sig_generate_sig_received) by (auto)
    then obtain X trecv l
      where       "(trecv, Recv (Rx B l) X) \<in> set tr" and
            sigX: "Crypt (priSK (Honest A)) msig \<in> subterms {X}" by auto
    with prems
      have "\<exists>D u tcsend Ld. (tcsend, Send (Tx D u) X Ld) \<in> set tr
            \<and> trecv >= cdist (Tx D u) (Rx B l) \<and> cdistM (Tx D u) (Rx B l) \<noteq> None
            \<and> tcsend \<le> trecv - cdist (Tx D u) (Rx B l)"
      by (intro send_before_recv) (auto)
    then obtain D tcsend u Ld
      where p1: "(tcsend, Send (Tx D u) X Ld) \<in> set tr" and
            p2: "trecv >= cdist (Tx D u) (Rx B l)" and
            p3: "cdistM (Tx D u) (Rx B l) \<noteq> None" and
            p4: "tcsend \<le> trecv - cdist (Tx D u) (Rx B l)" by auto
    let ?evd="(tcsend, Send (Tx D u) X Ld)"
    show ?thesis
    proof cases
      assume "Honest A=D"
      with p4 have p5: "cdist (Tx D u) (Rx B l) \<le> trecv - tcsend" by auto
      with `maxtime tr <= tsend` and `tsend=tb`
            and `(trecv, Recv (Rx B l) X) \<in> set tr`
      have tbgeq: "trecv <= tb" by (auto intro: maxtime_geq_elem)
      with p5 have p6:"cdist (Tx D u) (Rx B l) \<le> tb - tcsend" by auto
      with p3 have "cdistl D B \<le> cdist (Tx D u) (Rx B l)" apply -
	by (unfold cdist_def, rule noflt_some)
      with p6 `Honest A=D` have "cdistl (Honest A) B \<le> tb -tcsend" by auto
      with p1 sigX `Honest A=D`
        show ?thesis apply (rule_tac x=tcsend in exI, rule_tac x=X in exI)
        apply (rule_tac x=u in exI)
        by auto
    next
      assume "Honest A\<noteq>D"
      with prems have "\<exists>ta ma i La. cdistl (Honest A) D \<le> tcsend - ta \<and>
         (ta, Send (Tx (Honest A) i) ma La) \<in> set tr \<and>
         Crypt (priSK (Honest A)) msig \<in> subterms {ma}" by auto
      then obtain ta ma i La where cdistl: "cdistl (Honest A) D \<le> tcsend - ta" and
                                   intr: "(ta, Send (Tx (Honest A) i) ma La) \<in> set tr" and
                                   sig: "Crypt (priSK (Honest A)) msig \<in> subterms {ma}"
	by auto
      with p4 have p5: "cdistl (Honest A) D \<le>  trecv - cdist (Tx D u) (Rx B l) - ta"
	by auto
      show ?thesis
      proof (rule_tac x=ta in exI, rule_tac x=ma in exI, rule_tac x=i in exI, rule_tac x=La in exI,
	     intro conjI)
        from `maxtime tr <= tsend` and `tsend=tb`
             and `(trecv, Recv (Rx B l) X) \<in> set tr`
        have tr_tb: "trecv <= tb" by (auto intro: maxtime_geq_elem)
        from p3 noflt_some have "cdistl D B \<le> the (cdistM (Tx D u) (Rx B l))" by auto
        hence "cdistl D B \<le> cdist (Tx D u) (Rx B l)" by (auto simp add: cdist_def)
        with cdistl p5 have "cdistl (Honest A) D  \<le> trecv - ta - cdistl D B" by auto
        hence a1: "cdistl (Honest A) D + cdistl D B \<le> trecv - ta" by auto
        from cdistl_triangle
	  have "cdistl (Honest A) B \<le>  cdistl (Honest A) D + cdistl D B" by auto
        then have "cdistl (Honest A) B \<le> cdistl (Honest A) D + cdistl D B" by auto
        also with a1 have "\<dots> \<le> trecv - ta" by auto
        also with tr_tb have "\<dots> \<le> tb - ta" by auto
        finally show "cdistl (Honest A) B \<le> tb - ta" by auto
      next
        from prems show "(ta, Send (Tx (Honest A) i) ma La) \<in> set (?x # tr)" by auto
      next
        from prems show "Crypt (priSK (Honest A)) msig \<in> subterms {ma}" by auto
      qed
    qed
  next
    assume "?x \<noteq> ?evreply"
    with prems have "?evreply \<in> set tr" by auto
    with prems have "\<exists>ta ma i La. cdistl (Honest A) B \<le> tb - ta \<and>
       (ta, Send (Tx (Honest A) i) ma La) \<in> set tr \<and>
       Crypt (priSK (Honest A)) msig \<in> subterms {ma}"
      by auto
    thus ?case by auto
  qed
qed

lemma (in PROTOCOL_SYMKEYS_NOKEYS) mac_earliest_send:
  assumes sys_proto: "tr \<in> sys" and
          mcmac: "Hash (MPair (Key (symKey (Honest A) (Honest B))) mmac) \<in> subterms {mc}" and
          mcsent: "(tc, Send (Tx C j) mc Lc) \<in> set tr" and
          aneqc: "Honest A\<noteq>C" and
          bneqc: "Honest B\<noteq>C"
  shows "\<exists>te me i E Le. tc - te >= cdistl (Honest E) C \<and>
         (te, Send (Tx (Honest E) i) me Le) \<in> set tr
         \<and> (E = A \<or> E = B)
         \<and> (Hash (MPair (Key (symKey (Honest A) (Honest B))) mmac) \<in> subterms {me})"
  using sys_proto mcmac mcsent aneqc bneqc
proof (induct tr arbitrary: A B C j tc mc Lc mmac rule: sys.induct)
  case Nil thus ?case by auto

-- {* @{text "trace equal to: @{term (tc + tab, Recv D mc)#tr"} *}
next
  case (Con tr tc C k mc Lcc D l dCD trecv)
  hence "(tc, Send (Tx C j) mc Lc) \<in> set tr" by auto
  with Con.hyps prems show ?case by (auto)
-- {* @{text "(tsend, Send Intruder I mspy)#tr"} *}
next
  case (Fake tr tsend mintr I k)
  let ?key = "Key (symKey (Honest A) (Honest B))"
  let ?mac = "Hash (MPair ?key mmac)" and
      ?x = "(tsend, Send (Tx (Intruder I) k) mintr [])" and
      ?evreply = "(tc, Send (Tx C j) mc Lc)"

  show ?case
  proof cases
    assume xeq: "?x = ?evreply"
    with xeq have "C=Intruder I" and "mc=mintr" and "tsend=tc" by auto
    -- {* Sig in synth (Nonce I`.. u analz (knowsI I tr) => Recv X mit message
          => Send C mit message => IH*}
    with prems have "\<exists>trecv X i. (trecv, Recv (Rx C i) X) \<in> set tr
                     \<and> ?mac \<in> subterms {X}"
      apply -
      apply (rule mac_generate_mac_received)
      by auto
    then obtain X trecv l
      where       "(trecv, Recv (Rx C l) X) \<in> set tr" and
            macX: "?mac \<in> subterms {X}" by auto
    with prems
      have "\<exists>D u tdsend Ld. (tdsend, Send (Tx D u) X Ld) \<in> set tr
            \<and> trecv >= cdist (Tx D u) (Rx C l) \<and> cdistM (Tx D u) (Rx C l) \<noteq> None
            \<and> tdsend \<le> trecv - cdist (Tx D u) (Rx C l)"
      by (intro send_before_recv) (auto)
    then obtain D tdsend u Ld
      where p1: "(tdsend, Send (Tx D u) X Ld) \<in> set tr" and
            p2: "trecv >= cdist (Tx D u) (Rx C l)" and
            p3: "cdistM (Tx D u) (Rx C l) \<noteq> None" and
            p4: "tdsend \<le> trecv - cdist (Tx D u) (Rx C l)" by auto
    let ?evc="(tdsend, Send (Tx D u) X Ld)"
    show ?thesis
    proof cases
      assume "D \<in> {Honest A, Honest B}"
      from `D \<in> {Honest A, Honest B}` obtain E where "Honest E = D"
	by auto
      with p4 have p5: "cdist (Tx D u) (Rx C l) \<le> trecv - tdsend" by auto
      with `maxtime tr <= tsend` and `tsend=tc` and `(trecv, Recv (Rx C l) X) \<in> set tr`
        have tbgeq: "trecv <= tc" by (auto intro: maxtime_geq_elem)
      with p5 have p6:"cdist (Tx D u) (Rx C l) \<le> tc - tdsend" by auto
      with p3 have "cdistl D C \<le> cdist (Tx D u) (Rx C l)" apply -
	by (unfold cdist_def, rule noflt_some)
      with p6 `Honest E=D` have "cdistl (Honest E) C \<le> tc -tdsend" by auto
      with p1 macX `Honest E=D` `D \<in> {Honest A, Honest B}`
        show ?thesis apply (rule_tac x=tdsend in exI, rule_tac x=X in exI)
        apply (rule_tac x=u in exI, rule_tac x=E in exI)
	by auto
    next
      assume "D \<notin> {Honest A, Honest B}"
      with prems have "\<exists>te me i E Le. cdistl (Honest E) D \<le> tdsend - te \<and>
         (te, Send (Tx (Honest E) i) me Le) \<in> set tr \<and>
          (E = A \<or> E = B) \<and>
         (?mac) \<in> subterms {me}" by auto
      then obtain te me i E Le where cdistl: "cdistl (Honest E) D \<le> tdsend - te"
                               and intr: "(te, Send (Tx (Honest E) i) me Le) \<in> set tr"
                               and EisAB: "E = A \<or> E = B"
                               and mac: "(?mac) \<in> subterms {me}"
	                  by auto
      with p4 have p5: "cdistl (Honest E) D \<le>  trecv - cdist (Tx D u) (Rx C l) - te"
	by auto
      show ?thesis
      proof (rule_tac x=te in exI, rule_tac x=me in exI,
	     rule_tac x=i in exI, rule_tac x=E in exI,
             rule_tac x=Le in exI, intro conjI)
        from `maxtime tr <= tsend` and `tsend=tc` and
	     `(trecv, Recv (Rx C l) X) \<in> set tr`
          have tr_tb: "trecv <= tc" by (auto intro: maxtime_geq_elem)
        from p3 noflt_some have "cdistl D C \<le> the (cdistM (Tx D u) (Rx C l))" by auto
        hence "cdistl D C \<le> cdist (Tx D u) (Rx C l)" by (auto simp add: cdist_def)
        with cdistl p5 have "cdistl (Honest E) D  \<le> trecv - te - cdistl D C" by auto
        hence a1: "cdistl (Honest E) D + cdistl D C \<le> trecv - te" by auto
        from cdistl_triangle
	  have "cdistl (Honest E) C \<le>  cdistl (Honest E) D + cdistl D C" by auto
        then have "cdistl (Honest E) C \<le> cdistl (Honest E) D + cdistl D C" by auto
        also with a1 have "\<dots> \<le> trecv - te" by auto
        also with tr_tb have "\<dots> \<le> tc - te" by auto
        finally show "cdistl (Honest E) C \<le> tc - te" by auto
      next
        from prems show "(te, Send (Tx (Honest E) i) me Le) \<in> set (?x # tr)" by auto
      next
        from prems show "?mac \<in> subterms {me}" by auto
      next
	from prems show "E = A \<or> E = B" by auto
      qed
    qed
  next
    assume "?x \<noteq> ?evreply"
    with prems have "?evreply \<in> set tr" by auto
    with prems have "\<exists>te me i E Le. cdistl (Honest E) C \<le> tc - te \<and>
       (te, Send (Tx (Honest E) i) me Le) \<in> set tr \<and>
       (E = A \<or> E = B) \<and>
       ?mac \<in> subterms {me}" by auto
    thus ?case by auto
  qed
next
  case (Proto tr tsend step m pEv D)
  let ?key = "Key (symKey (Honest A) (Honest B))"
  let ?mac = "Hash (MPair ?key mmac)" and
      ?x = "(tsend, createEv D pEv m)" and
      ?evreply = "(tc, Send (Tx C j) mc Lc)"

  show ?case
  proof cases
    assume xeq: "?x = ?evreply"
    with xeq have "C=Honest D" and "mc=m" and "tsend=tc" apply auto
      apply (case_tac "pEv") apply auto apply (case_tac "pEv")
      by auto
    with prems have "\<exists>trecv X i. (trecv, Recv (Rx C i) X) \<in> set tr
                     \<and> ?mac \<in> subterms {X}"
      apply - apply (rule mac_generate_mac_received) by auto
    then obtain X trecv l
      where       "(trecv, Recv (Rx C l) X) \<in> set tr" and
            macX: "?mac \<in> subterms {X}" by auto
    with prems
      have "\<exists>F u tfsend Lf. (tfsend, Send (Tx F u) X Lf) \<in> set tr
            \<and> trecv >= cdist (Tx F u) (Rx C l) \<and> cdistM (Tx F u) (Rx C l) \<noteq> None
            \<and> tfsend \<le> trecv - cdist (Tx F u) (Rx C l)"
      by (intro send_before_recv) (auto)
    then obtain F tfsend u Lf
      where p1: "(tfsend, Send (Tx F u) X Lf) \<in> set tr" and
            p2: "trecv >= cdist (Tx F u) (Rx C l)" and
            p3: "cdistM (Tx F u) (Rx C l) \<noteq> None" and
            p4: "tfsend \<le> trecv - cdist (Tx F u) (Rx C l)" by auto
    let ?evd="(tfsend, Send (Tx F u) X Lf)"
    show ?thesis
    proof cases
      assume "F \<in> {Honest A, Honest B}"
      then obtain E where "Honest E=F" by auto
      with p4 have p5: "cdist (Tx F u) (Rx C l) \<le> trecv - tfsend" by auto
      with `maxtime tr <= tsend` and `tsend=tc`
            and `(trecv, Recv (Rx C l) X) \<in> set tr`
      have tbgeq: "trecv <= tc" by (auto intro: maxtime_geq_elem)
      with p5 have p6:"cdist (Tx F u) (Rx C l) \<le> tc - tfsend" by auto
      with p3 have "cdistl F C \<le> cdist (Tx F u) (Rx C l)" apply -
	by (unfold cdist_def, rule noflt_some)
      with p6 `Honest E=F` have "cdistl (Honest E) C \<le> tc -tfsend" by auto
      with p1 macX `Honest E=F` `F \<in> {Honest A, Honest B}`
        show ?thesis apply (rule_tac x=tfsend in exI, rule_tac x=X in exI)
        apply (rule_tac x=u in exI, rule_tac x=E in exI)
        by auto
    next
      assume "F \<notin> {Honest A, Honest B}"
      with prems have "\<exists>te me i E Le. cdistl (Honest E) F \<le> tfsend - te \<and>
         (te, Send (Tx (Honest E) i) me Le) \<in> set tr \<and>
         (E = A \<or> E = B) \<and>
         ?mac \<in> subterms {me}" by auto
      then obtain te me i E Le where cdistl: "cdistl (Honest E) F \<le> tfsend - te" and
	                             EisAB: "(E = A \<or> E = B)" and
                                     intr: "(te, Send (Tx (Honest E) i) me Le) \<in> set tr" and
                                     mac: "?mac \<in> subterms {me}"
	by auto
      with p4 have p5: "cdistl (Honest E) F \<le>  trecv - cdist (Tx F u) (Rx C l) - te"
	by auto
      show ?thesis
      proof (rule_tac x=te in exI, rule_tac x=me in exI, rule_tac x=i in exI, rule_tac x=E in exI,
	     rule_tac x=Le in exI, intro conjI)
        from `maxtime tr <= tsend` and `tsend=tc`
             and `(trecv, Recv (Rx C l) X) \<in> set tr`
        have tr_tb: "trecv <= tc" by (auto intro: maxtime_geq_elem)
        from p3 noflt_some have "cdistl F C \<le> the (cdistM (Tx F u) (Rx C l))" by auto
        hence "cdistl F C \<le> cdist (Tx F u) (Rx C l)" by (auto simp add: cdist_def)
        with cdistl p5 have "cdistl (Honest E) F  \<le> trecv - te - cdistl F C" by auto
        hence a1: "cdistl (Honest E) F + cdistl F C \<le> trecv - te" by auto
        from cdistl_triangle
	  have "cdistl (Honest E) C \<le>  cdistl (Honest E) F + cdistl F C" by auto
        then have "cdistl (Honest E) C \<le> cdistl (Honest E) F + cdistl F C" by auto
        also with a1 have "\<dots> \<le> trecv - te" by auto
        also with tr_tb have "\<dots> \<le> tc - te" by auto
        finally show "cdistl (Honest E) C \<le> tc - te" by auto
      next
        from prems show "(te, Send (Tx (Honest E) i) me Le) \<in> set (?x # tr)" by auto
      next
        from prems show "?mac \<in> subterms {me}" by auto
      next
	from prems show "E=A \<or> E =B" by auto
      qed
    qed
  next
    assume "?x \<noteq> ?evreply"
    with prems have "?evreply \<in> set tr" by auto
    with prems have "\<exists>te me i E Le. cdistl (Honest E) C \<le> tc - te \<and>
       (te, Send (Tx (Honest E) i) me Le) \<in> set tr \<and>
       (E=A \<or> E =B) \<and> 
       ?mac \<in> subterms {me}"
      by auto
    thus ?case
      apply (elim exE)
      apply (rule_tac x=te in exI)
      by auto
  qed
qed

lemma (in PROTOCOL_SYMKEYS_NOKEYS) mac_originates:
  assumes sys_proto: "tr \<in> sys" and
          mcmac: "Hash (MPair (Key (symKey (Honest A) (Honest B))) mmac) \<in> subterms {mc}" and
          mcsent: "(tc, Send (Tx C j) mc Lc) \<in> set tr"
  shows "\<exists>te me i E Le.
         (te, Send (Tx (Honest E) i) me Le) \<in> set tr
         \<and> (E = A \<or> E = B)
         \<and> (Hash (MPair (Key (symKey (Honest A) (Honest B))) mmac) \<in> subterms {me})
         \<and> (Hash (MPair (Key (symKey (Honest A) (Honest B))) mmac)
            \<notin> used (beforeEvent (te, Send (Tx (Honest E) i) me Le) tr))"
  using sys_proto mcmac mcsent
proof (induct tr arbitrary: A B C j tc mc mmac Lc rule: sys.induct)
  case Nil thus ?case by auto

-- {* @{text "trace equal to: @{term (tc + tab, Recv D mc)#tr"} *}
next
  case (Con tr tc C k mc D l dCD trecv)
  hence "(tc, Send (Tx C j) mc Lc) \<in> set tr" by auto
  with Con.hyps prems show ?case by auto
-- {* @{text "(tsend, Send Intruder I mspy)#tr"} *}
next
  case (Fake tr tsend mintr I k)
  let ?key = "Key (symKey (Honest A) (Honest B))"
  let ?mac = "Hash (MPair ?key mmac)" and
      ?x = "(tsend, Send (Tx (Intruder I) k) mintr [])" and
      ?evreply = "(tc, Send (Tx C j) mc Lc)"

  show ?case
  proof cases
    assume xeq: "?x = ?evreply"
    with xeq have "C=Intruder I" and "mc=mintr" and "tsend=tc" by auto
    -- {* MAC in synth (Nonce I`.. u analz (knowsI I tr) => Recv X mit message
          => Send C mit message => IH*}
    with prems have "\<exists>trecv X i. (trecv, Recv (Rx C i) X) \<in> set tr
                     \<and> ?mac \<in> subterms {X}" apply -
      apply (rule mac_generate_mac_received)
      by auto
    then obtain X trecv l
      where ctr:  "(trecv, Recv (Rx C l) X) \<in> set tr" and
            macX: "?mac \<in> subterms {X}" by auto
    with prems
      have "\<exists>D u tdsend Ld. (tdsend, Send (Tx D u) X Ld) \<in> set tr
            \<and> trecv >= cdist (Tx D u) (Rx C l) \<and> cdistM (Tx D u) (Rx C l) \<noteq> None
            \<and> tdsend \<le> trecv - cdist (Tx D u) (Rx C l)"
      by (intro send_before_recv) (auto)
    then obtain D tdsend u Ld
      where p1: "(tdsend, Send (Tx D u) X Ld) \<in> set tr" and
            p2: "trecv >= cdist (Tx D u) (Rx C l)" and
            p3: "cdistM (Tx D u) (Rx C l) \<noteq> None" and
            p4: "tdsend \<le> trecv - cdist (Tx D u) (Rx C l)" by auto
    from `maxtime tr <= tsend` and `tsend=tc` and
  	     `(trecv, Recv (Rx C l) X) \<in> set tr`
          have tr_tb: "trecv <= tc" by (auto intro: maxtime_geq_elem)
    thus ?thesis proof cases
      assume DAB: "D \<in> {Honest A, Honest B}"
      obtain HD where D: "D = Honest HD" using DAB by auto
      thus ?thesis using prems DAB apply -
	apply (drule_tac C="D" in Fake.hyps(2)) back back
	apply force
	apply auto
	done
    next
      assume nAB: "D \<notin> {Honest A, Honest B}"
      from p1 p2 p3 p4 prems have "\<exists>te me i E Le.
           (te, Send (Tx (Honest E) i) me Le) \<in> set tr \<and>
           (E = A \<or> E = B) \<and>
           (?mac) \<in> subterms {me}
           \<and> (?mac \<notin> used (beforeEvent (te, Send (Tx (Honest E) i) me Le) tr))" apply -
	apply (rule Fake.hyps(2))
	apply auto
	done
      then obtain te me i E Le where intr: "(te, Send (Tx (Honest E) i) me Le) \<in> set tr"
                               and EisAB: "E = A \<or> E = B"
                               and mac: "(?mac) \<in> subterms {me}"
	                       and fresh: "(?mac \<notin> used (beforeEvent (te, Send (Tx (Honest E) i) me Le) tr))"
        by auto
      show ?thesis
      proof (rule_tac x=te in exI, rule_tac x=me in exI,
  	     rule_tac x=i in exI, rule_tac x=E in exI, rule_tac x=Le in exI, intro conjI)
	from prems show "(te, Send (Tx (Honest E) i) me Le) \<in> set (?x # tr)" by auto
      next
	from prems show "?mac \<in> subterms {me}" by auto
      next
	from prems show "E = A \<or> E = B" by auto
      next
	from prems show "?mac \<notin> used (beforeEvent (te, Send (Tx (Honest E) i) me Le)
                                     ((tsend, Send (Tx (Intruder I) k) mintr [])#tr))"
	  by auto
      qed
    qed
  next
    assume "?x \<noteq> ?evreply"
    with prems have "?evreply \<in> set tr" by auto
    thus ?case using prems by auto
  qed
next
  case (Proto tr tsend step m pEv D)
  let ?key = "Key (symKey (Honest A) (Honest B))"
  let ?mac = "Hash (MPair ?key mmac)" and 
      ?x = "(tsend, createEv D pEv m)" and
      ?evreply = "(tc, Send (Tx C j) mc Lc)"
  
  show ?case proof cases
    assume xeq: "?x = ?evreply"
    with xeq prems have Ceq: "C=Honest D" and meq: "mc=m" and tceq: "tsend=tc" apply auto
      apply (case_tac pEv, auto)+
      done
    -- {* MAC in synth (Nonce I`.. u analz (knowsI I tr) => Recv X mit message
          => Send C mit message => IH*}
    show ?thesis proof cases
      assume "\<exists> te me i E Le. (te, Send (Tx E i) me Le) \<in> set tr \<and>
                           Hash (MPair (Key (symKey (Honest A) (Honest B))) mmac) \<in> subterms {me}"
      thus ?thesis using prems Ceq meq tceq apply -
	apply (elim exE conjE)
	apply (drule Proto.hyps(2)) back back
	apply simp
	apply (elim exE conjE)
	apply (rule_tac x=teb in exI) apply (rule_tac x=meb in exI)
	apply (rule_tac x=ib in exI) apply (rule_tac x=Eb in exI)
	apply (rule_tac x=Leb in exI)
	apply (intro conjI)
	apply force
	apply force
	apply force
	apply force
	done
      next
	assume notex: "\<not> (\<exists> te me i E Le. (te, Send (Tx E i) me Le) \<in> set tr \<and>
                          Hash (MPair (Key (symKey (Honest A) (Honest B))) mmac) \<in> subterms {me})"
	hence fresh: "Hash (MPair (Key (symKey (Honest A) (Honest B))) mmac)
                      \<notin> used (beforeEvent ?x (?x # tr))"
	proof cases
	  assume "?x \<in> set tr"
	  thus ?thesis using notex xeq
	    apply auto
	    apply (drule Used_imp_send_parts)
	    apply auto
	    apply (drule beforeEvent_subset)
	    apply (case_tac A,auto)
	    apply (case_tac pEv, auto)
	    apply force
	    done
	next
	  assume "?x \<notin> set tr"
	  thus ?thesis using notex
	    apply auto
	    apply (drule Used_imp_send_parts)
	    apply auto
	    apply (case_tac A, auto)
	    apply force
	    done
	qed
	show ?thesis proof cases
	  assume DAB: "D \<in> {A,B}"
	  thus ?thesis using prems xeq Ceq fresh
	    apply (rule_tac x=tc in exI)
	    apply (rule_tac x=mc in exI)
	    apply (rule_tac x=j in exI)
	    apply (rule_tac x="D" in exI)
	    apply (rule_tac x="Lc" in exI)
	    apply (intro conjI)
	    apply force
	    apply force
	    apply force
	    apply force
	    done
	next
	 assume nDAB: "D \<notin> {A,B}"
	 show ?thesis using prems apply -
	   apply (frule_tac tr=tr and m=m and B="Honest D" in mac_generate_mac_received)
	   apply force
	   apply (force simp add: meq)
	   apply force
	   apply (elim exE)
	   apply (drule send_before_recv) back
	   apply force
	   apply (elim exE conjE)
	   apply auto
	   apply (erule_tac x=tsend in allE)
	   apply (erule_tac x=X in allE)
	   apply (erule disjE)
	   apply auto
	   done
       qed
     qed
  next
    assume "?x \<noteq> ?evreply"
    with prems have "?evreply \<in> set tr" by auto
    thus ?thesis using prems apply -
      apply (drule_tac tc=tc and C=C and j=j and mc=mc in Proto.hyps(2))
      apply assumption
      apply (elim exE conjE)
      apply (rule_tac x=te in exI) apply (rule_tac x=me in exI)
      apply (rule_tac x=i in exI)  apply (rule_tac x=E in exI)
      by auto
  qed
qed

lemma (in PROTOCOL_NONONCE) fresh_nonce_earliest_recv:
  assumes sys_proto: "tr \<in> sys" and
          fresh: "Nonce A NA
                  \<notin> used (beforeEvent (ta, Send (Tx A i) ma La) tr)" and
          manonce: "Nonce A NA \<in> subterms {ma}" and
          mbnonce: "Nonce A NA \<in> subterms {mb}" and
          masend:  "(ta, Send (Tx A i) ma La) \<in> set tr" and
          mbrecv:  "(tb, Recv (Rx B j) mb) \<in> set tr" and
          aneqb: "A\<noteq>B"
  shows "tb - ta >= cdistl A B"
  using sys_proto fresh manonce mbnonce masend mbrecv aneqb
proof (induct tr arbitrary: A B ta tb ma mb La i j NA rule: sys.induct) 
  case Nil thus ?case by auto

-- {* @{text "(tsend, Send Intruder I mspy)#tr"} *}
next
  case (Fake tr  tsend mspy I k)
  let ?x = "(tsend, Send (Tx (Intruder I) k) mspy [])" and
      ?eva = "(ta, Send (Tx A i) ma La)" and ?evb = "(tb, Recv (Rx B j) mb)"
  note caserule =  set_two_elem_cases [where eva="?eva" and evb="?evb"
                                       and tr="tr" and x="?x"]
  have evbneqx: "?evb\<noteq>?x" by auto 
  from prems show ?case
  proof (cases rule: caserule, simp, simp)
    case 3
    note `?eva \<in> set tr` and `?evb \<in> set tr`
    show ?thesis
    proof cases
      assume "?eva=?x \<and> ?eva \<notin> set tr"
      with prems
      have "\<exists>C k tcsend Lb. (tcsend, Send (Tx C k) mb Lb) \<in> set tr
            \<and> tb >= cdist (Tx C k) (Rx B j)
            \<and> cdistM (Tx C k) (Rx B j) \<noteq> None
            \<and> tcsend \<le> tb - cdist (Tx C k) (Rx B j)"
      by (intro send_before_recv) (auto)
      then obtain C k tcsend Lc where
        p1:"(tcsend, Send (Tx C k) mb Lc) \<in> set tr" by auto
      from `?eva=?x \<and> ?eva \<notin> set tr` `Nonce A NA \<notin> used (beforeEvent ?eva (?x#tr))` have
        p5: "Nonce A NA \<notin> used tr" by auto
      from p1 and `Nonce A NA \<in> subterms {mb}` have "Nonce A NA \<in> used tr"
        apply - apply (rule Send_imp_parts_used) by auto
      with p5 show ?thesis by contradiction
    next
      assume "\<not> (?eva=?x \<and> ?eva \<notin> set tr)"
      with prems show ?thesis by auto
    qed
  next
    case 6
    note `?eva = ?x` and `?evb = ?x`
    with evbneqx show ?thesis by auto
  next
    case 4
    note `?eva \<in> set tr` and beq=`?evb = ?x` and `?eva \<noteq> ?x`
    with evbneqx show ?thesis by auto
  next
    case 5
    note `?evb \<in> set tr` and `?eva = ?x` and `?evb \<noteq> ?x`
    with prems
      have "\<exists>C k tcsend Lc. (tcsend, Send (Tx C k) mb Lc) \<in> set tr
            \<and> tb >= cdist (Tx C k) (Rx B j)
            \<and> cdistM (Tx C k) (Rx B j) \<noteq> None
            \<and> tcsend \<le> tb - cdist (Tx C k) (Rx B j)"
      by (intro send_before_recv) (auto)
    then obtain C k tcsend Lc where
      p1:"(tcsend, Send (Tx C k) mb Lc) \<in> set tr" by auto
    show ?thesis proof cases
      assume "?eva \<in> set tr"
      thus ?thesis using `?evb \<in> set tr``?eva=?x` prems apply -
	apply (rule Fake.hyps(2)) prefer 5
	apply assumption defer
	apply auto
	done
    next
      assume "?eva \<notin> set tr"
      with `?eva=?x` `Nonce A NA \<notin> used (beforeEvent ?eva (?x#tr))` have
	p5: "Nonce A NA \<notin> used tr" by auto
      from p1 and `Nonce A NA \<in> subterms {mb}` have "Nonce A NA \<in> used tr"
	apply - apply (rule Send_imp_parts_used) by auto
      with p5 show ?thesis by contradiction
    qed
  qed

next
  case (Proto tr tsend step m pEv C)
  let ?x = "(tsend, createEv C pEv m)" and
      ?eva = "(ta, Send (Tx A i) ma La)" and ?evb = "(tb, Recv (Rx B j) mb)"
  note caserule =  set_two_elem_cases [where eva="?eva" and evb="?evb"
                                       and tr="tr" and x="?x"]
  have evbneqx: "?evb\<noteq>?x" by auto
  from prems show ?case
  proof (cases rule: caserule, simp, simp)
    case 3
    note `?eva \<in> set tr` and `?evb \<in> set tr`
    show ?thesis
    proof cases
      assume n: "?eva=?x \<and> ?eva \<notin> set tr"
      with prems
      have "\<exists>C k tcsend Lc. (tcsend, Send (Tx C k) mb Lc) \<in> set tr
            \<and> tb >= cdist (Tx C k) (Rx B j)
            \<and> cdistM (Tx C k) (Rx B j) \<noteq> None
            \<and> tcsend \<le> tb - cdist (Tx C k) (Rx B j)"
      by (intro send_before_recv) (auto)
      then obtain C k tcsend Lc where
        p1:"(tcsend, Send (Tx C k) mb Lc) \<in> set tr" by auto
      from n `Nonce A NA \<notin> used (beforeEvent ?eva (?x#tr))` have
        p5: "Nonce A NA \<notin> used tr" by auto
      from n p1 `Nonce A NA \<in> subterms {mb}` have "Nonce A NA \<in> used tr"
        apply - apply (rule Send_imp_parts_used) by auto
      with p5 show ?thesis by contradiction
    next
      assume "\<not> (?eva=?x \<and> ?eva \<notin> set tr)"
      with prems show ?thesis by auto
    qed
  next
    case 6
    note `?eva = ?x` and `?evb = ?x`
    with evbneqx show ?thesis by auto
  next
    case 4
    note `?eva \<in> set tr` and beq=`?evb = ?x` and `?eva \<noteq> ?x`
    with evbneqx show ?thesis by auto
  next
    case 5
    note `?evb \<in> set tr` and `?eva = ?x` and `?evb \<noteq> ?x`
    with prems
      have "\<exists>C k tcsend Lc. (tcsend, Send (Tx C k) mb Lc) \<in> set tr
            \<and> tb >= cdist (Tx C k) (Rx B j)
            \<and> cdistM (Tx C k) (Rx B j) \<noteq> None
            \<and> tcsend \<le> tb - cdist (Tx C k) (Rx B j)"
      by (intro send_before_recv) (auto)
    then obtain C k tcsend Lc where
      p1:"(tcsend, Send (Tx C k) mb Lc) \<in> set tr" by auto
    show ?thesis proof cases
      assume "?eva \<in> set tr"
      thus ?thesis using prems apply -
	apply (rule Proto.hyps(2)) prefer 5
	apply assumption+ defer
	apply assumption+
	apply simp
	done
    next
      assume "?eva \<notin> set tr"
      with `?eva=?x` `Nonce A NA \<notin> used (beforeEvent ?eva (?x#tr))` have
	p5: "Nonce A NA \<notin> used tr" by auto
      from p1 and `Nonce A NA \<in> subterms {mb}` have "Nonce A NA \<in> used tr"
	apply - apply (rule Send_imp_parts_used) by auto
      with p5 show ?thesis by contradiction
    qed
  qed

-- {* @{text "trace equal to: @{term (tc + tab, Recv D mc)#tr"} *}
next
  case  (Con tr tcrecv tsend C k Y Ld D l dCD)
  let ?x = "(tcrecv, Recv (Rx D l) Y)" and
      ?eva = "(ta, Send (Tx A i) ma La)" and ?evb = "(tb, Recv (Rx B j) mb)"
  note caserule =  set_two_elem_cases [where eva="?eva" and evb="?evb"
                                       and tr="tr" and x="?x"]
  have evaneqx: "?eva\<noteq>?x" by auto
  from prems show ?case
  proof (cases rule: caserule, simp, simp)
    case 6
    note `?eva = ?x` and `?evb = ?x`
    with prems evaneqx show ?thesis by auto
  next
    case 5
    note `?evb \<in> set tr` and `?eva = ?x` and `?evb \<noteq> ?x`
    with prems evaneqx show ?thesis by auto
  next
    case 3
    note `?eva \<in> set tr` and `?evb \<in> set tr`
    with `Nonce A NA \<notin> used (beforeEvent ?eva (?x#tr))`
      have "Nonce A NA \<notin> used (beforeEvent ?eva tr)" by auto
    with prems show ?thesis by auto
  next
    case 4
    note `?eva \<in> set tr` and beq=`?evb = ?x` and `?eva \<noteq> ?x`
    note p1=`(tsend, Send (Tx C k) Y Ld) \<in> set tr` and p2=`tsend + dCD \<le> tcrecv` and
         p3=`cdistM (Tx C k) (Rx D l) = Some dCD`
    let ?evc="(tsend, Send (Tx C k) mb Ld)"
    from beq have eq1: "Y=mb" and eq2:"D=B" and eq3:"tcrecv=tb" and eq4:"l=j" by auto
    with p1 have p1: "?evc \<in> set tr" by auto
    show ?thesis
    proof cases
      assume "C=A"
      with prems `?eva\<noteq>?x` have nafresh: "Nonce A NA \<notin> used (beforeEvent ?eva tr)"
        by auto
      with p1 `Nonce A NA \<in> subterms {mb}` `C=A` have "?evc \<notin> set (beforeEvent ?eva tr)"
        apply - apply (rule nonce_orig_not_before) by auto
      with prems p1 `C=A` have p5: "tsend \<ge> ta"
        apply - apply (rule not_beforeEvent_later) by auto
      from p3 p2 eq1 eq2 eq3 eq4 `C=A` have "cdist (Tx A k) (Rx B j) \<le> tb - tsend"
	by (simp add: cdist_def)
      with p5 also have p6: "cdist (Tx A k) (Rx B j) \<le> tb - ta" by auto
      with p3 eq2 eq4 `C=A` have "cdistl A B \<le> cdist (Tx A k) (Rx B j)"
        apply - apply (unfold cdist_def, rule noflt_some) by simp
      with p6 show ?thesis by auto
    next
      assume "C\<noteq>A"
      hence p4: "cdistl A C \<le> tsend - ta" using prems
        by (auto intro: fresh_nonce_earliest_send)
      from beq have eq1: "tb=tcrecv" and eq2: "B=D" and eq3: "j=l" by auto
      with p3 have p5: "cdistl C B \<le> cdist (Tx C k) (Rx B j)" apply -
        apply (unfold cdist_def, rule noflt_some) by auto
      with p2 p3 eq2 eq3 eq1 have p6: "tb - tsend \<ge> cdist (Tx C k) (Rx B j)"
	by (auto simp add: cdist_def)
      with cdistl_triangle have "cdistl A B \<le> cdistl A C + cdistl C B"
	by (auto simp add: cdist_def)
      also with p5 have "\<dots> \<le> cdistl A C + cdist (Tx C k) (Rx B j)" by auto
      also with p4 p6 have "\<dots> \<le> tb - ta" by arith
      finally show ?thesis .
    qed
  qed
qed

lemma (in PROTOCOL_NONONCE) nonce_usedI_view:
  "[| Nonce (Honest A) NA \<in> usedI tr; tr \<in> sys |]
  ==> Nonce (Honest A) NA \<in> usedI (view A tr)"
  apply (auto simp add: usedI_def)
  apply (rule nonce_used_view)
  apply (auto)
done

lemma (in PROTOCOL_NONONCE) nonce_view_fresh:
  "tr \<in> sys \<Longrightarrow> 
   (Nonce (Honest A) NA \<notin> usedI (view A tr)) =
   (Nonce (Honest A) NA \<notin> usedI tr)"
  apply auto prefer 2
  apply (erule nonce_usedI_view, simp)
  apply (rule usedI_mono_snd) prefer 2
  apply force
  apply (auto simp add: view_def split: split_if_asm)
  apply (rule image_eqI)
  apply auto
done

lemma (in PROTOCOL_NONONCE) nonce_view_used:
  "tr \<in> sys \<Longrightarrow> 
   (Nonce (Honest A) NA \<in> usedI (view A tr)) =
   (Nonce (Honest A) NA \<in> usedI tr)"
  apply auto prefer 2
  apply (erule nonce_usedI_view, simp)
  apply (rule usedI_mono_snd) prefer 2
  apply force
  apply (auto simp add: view_def split: split_if_asm)
  apply (rule image_eqI)
  apply auto
done

(* MOVE to SystemOrigination*)
lemma (in MESSAGE_DERIVATION) originate_unique:
  assumes "m \<notin> used (beforeEvent (ta, Send TA ma La) tr)"
  and     "m \<notin> used (beforeEvent (tb, Send TB mb Lb) tr)"
  and     "(tb, Send TB mb Lb) \<noteq> (ta, Send TA ma La)"
  and     "(tb, Send TB mb Lb) \<in> set tr"
  and     "(ta, Send TA ma La) \<in> set tr"
  and     " m \<in> subterms {ma}"
  shows   "m \<notin> subterms {mb}" using prems
  apply (induct tr)
  apply simp
  apply (case_tac "a=(ta, Send TA ma La) \<and> a \<notin> set tr")
  apply (elim conjE)
  apply simp
  apply (case_tac "m \<in> subterms {mb}") prefer 2
  apply force
  apply (subgoal_tac "(tb, Send TB mb Lb) \<in> set tr") prefer 2
  apply force
  apply (frule_tac Y=m in Send_imp_parts_used)
  apply force
  apply force
  apply (case_tac "a=(tb, Send TB mb Lb) \<and> a \<notin> set tr")
  apply (elim conjE)
  apply simp
  apply (subgoal_tac "(ta, Send TA ma La) \<in> set tr") prefer 2
  apply force
  apply (frule_tac Y=m in Send_imp_parts_used)
  apply force
  apply force
  apply auto
done

end
